import meep as mp

try:
    import meep.adjoint as mpa
except:
    import adjoint as mpa

import unittest
from enum import Enum

import numpy as np
from autograd import numpy as npa
from autograd import tensor_jacobian_product
from utils import ApproxComparisonTestCase

MonitorObject = Enum("MonitorObject", "EIGENMODE DFT LDOS")


class TestAdjointSolver(ApproxComparisonTestCase):
    @classmethod
    def setUpClass(cls):
        cls.resolution = 30  # pixels/μm

        cls.silicon = mp.Medium(epsilon=12)
        cls.sapphire = mp.Medium(
            epsilon_diag=(10.225, 10.225, 9.95),
            epsilon_offdiag=(-0.825, -0.55 * np.sqrt(3 / 2), 0.55 * np.sqrt(3 / 2)),
        )

        cls.sxy = 5.0
        cls.cell_size = mp.Vector3(cls.sxy, cls.sxy, 0)

        cls.dpml = 1.0
        cls.pml_xy = [mp.PML(thickness=cls.dpml)]
        cls.pml_x = [mp.PML(thickness=cls.dpml, direction=mp.X)]

        cls.eig_parity = mp.EVEN_Y + mp.ODD_Z

        cls.design_region_size = mp.Vector3(1.5, 1.5)
        cls.design_region_resolution = int(2 * cls.resolution)
        cls.Nx = int(cls.design_region_size.x * cls.design_region_resolution)
        cls.Ny = int(cls.design_region_size.y * cls.design_region_resolution)

        # ensure reproducible results
        rng = np.random.RandomState(9861548)

        # random design region
        cls.p = 0.5 * rng.rand(cls.Nx * cls.Ny)

        # random perturbation for design region
        deps = 1e-5
        cls.dp = deps * rng.rand(cls.Nx * cls.Ny)

        cls.w = 1.0
        cls.waveguide_geometry = [
            mp.Block(
                material=cls.silicon,
                center=mp.Vector3(),
                size=mp.Vector3(mp.inf, cls.w, mp.inf),
            )
        ]

        cls.fcen = 1 / 1.55
        cls.df = 0.2 * cls.fcen
        cls.mode_source = [
            mp.EigenModeSource(
                src=mp.GaussianSource(cls.fcen, fwidth=cls.df),
                center=mp.Vector3(-0.5 * cls.sxy + cls.dpml, 0),
                size=mp.Vector3(0, cls.sxy - 2 * cls.dpml),
                eig_parity=cls.eig_parity,
            )
        ]

        cls.pt_source = [
            mp.Source(
                src=mp.GaussianSource(cls.fcen, fwidth=cls.df),
                center=mp.Vector3(-0.5 * cls.sxy + cls.dpml, 0),
                size=mp.Vector3(),
                component=mp.Ez,
            )
        ]

        cls.line_source = [
            mp.Source(
                src=mp.GaussianSource(cls.fcen, fwidth=cls.df),
                center=mp.Vector3(-0.85, 0),
                size=mp.Vector3(0, cls.sxy - 2 * cls.dpml),
                component=mp.Ez,
            )
        ]

        cls.k_point = mp.Vector3(0.23, -0.38)

        # location of DFT monitors for reflected and transmitted fields
        cls.refl_pt = mp.Vector3(-0.5 * cls.sxy + cls.dpml + 0.5, 0)
        cls.tran_pt = mp.Vector3(0.5 * cls.sxy - cls.dpml, 0)

    def adjoint_solver(
        self,
        design_params,
        mon_type: MonitorObject,
        frequencies=None,
        mat2=None,
        need_gradient=True,
    ):
        matgrid = mp.MaterialGrid(
            mp.Vector3(self.Nx, self.Ny),
            mp.air,
            self.silicon if mat2 is None else mat2,
            weights=np.ones((self.Nx, self.Ny)),
        )

        matgrid_region = mpa.DesignRegion(
            matgrid,
            volume=mp.Volume(
                center=mp.Vector3(),
                size=mp.Vector3(
                    self.design_region_size.x, self.design_region_size.y, 0
                ),
            ),
        )

        matgrid_geometry = [
            mp.Block(
                center=matgrid_region.center, size=matgrid_region.size, material=matgrid
            )
        ]

        geometry = self.waveguide_geometry + matgrid_geometry

        sim = mp.Simulation(
            resolution=self.resolution,
            cell_size=self.cell_size,
            boundary_layers=self.pml_xy,
            sources=self.mode_source,
            geometry=geometry,
        )

        if not frequencies:
            frequencies = [self.fcen]

        if mon_type.name == "EIGENMODE":
            if len(frequencies) == 1:
                if mat2 is None:
                    # Compute the incident fields of the mode source
                    # in the straight waveguide for use as normalization
                    # of the reflectance (S11) measurement.
                    ref_sim = mp.Simulation(
                        resolution=self.resolution,
                        cell_size=self.cell_size,
                        boundary_layers=self.pml_xy,
                        sources=self.mode_source,
                        geometry=self.waveguide_geometry,
                    )
                    dft_mon = ref_sim.add_mode_monitor(
                        frequencies,
                        mp.ModeRegion(
                            center=self.refl_pt,
                            size=mp.Vector3(0, self.sxy - 2 * self.dpml, 0),
                        ),
                        yee_grid=True,
                    )
                    ref_sim.run(until_after_sources=20)
                    subtracted_dft_fields = ref_sim.get_flux_data(dft_mon)
                else:
                    subtracted_dft_fields = None

                obj_list = [
                    mpa.EigenmodeCoefficient(
                        sim,
                        mp.Volume(
                            center=self.refl_pt,
                            size=mp.Vector3(0, self.sxy - 2 * self.dpml, 0),
                        ),
                        1,
                        forward=False,
                        eig_parity=self.eig_parity,
                        subtracted_dft_fields=subtracted_dft_fields,
                    ),
                    mpa.EigenmodeCoefficient(
                        sim,
                        mp.Volume(
                            center=self.tran_pt,
                            size=mp.Vector3(0, self.sxy - 2 * self.dpml, 0),
                        ),
                        2,
                        eig_parity=self.eig_parity,
                    ),
                ]

                def J(refl_mon, tran_mon):
                    return -npa.power(npa.abs(refl_mon), 2) + npa.power(
                        npa.abs(tran_mon), 2
                    )

            else:
                obj_list = [
                    mpa.EigenmodeCoefficient(
                        sim,
                        mp.Volume(
                            center=mp.Vector3(0.5 * self.sxy - self.dpml),
                            size=mp.Vector3(0, self.sxy - 2 * self.dpml, 0),
                        ),
                        1,
                        eig_parity=self.eig_parity,
                    )
                ]

                def J(tran_mon):
                    return npa.power(npa.abs(tran_mon), 2)

        elif mon_type.name == "DFT":
            obj_list = [
                mpa.FourierFields(
                    sim,
                    mp.Volume(center=mp.Vector3(1.25), size=mp.Vector3(0.25, 1, 0)),
                    mp.Ez,
                )
            ]

            def J(mode_mon):
                return npa.power(npa.abs(mode_mon[:, 4, 10]), 2)

        elif mon_type.name == "LDOS":
            sim.change_sources(self.line_source)
            obj_list = [mpa.LDOS(sim)]

            def J(mode_mon):
                return npa.array(mode_mon)

        opt = mpa.OptimizationProblem(
            simulation=sim,
            objective_functions=J,
            objective_arguments=obj_list,
            design_regions=[matgrid_region],
            frequencies=frequencies,
        )

        if need_gradient:
            f, dJ_du = opt([design_params])
            return f, dJ_du
        else:
            f = opt([design_params], need_gradient=False)
            return f[0]

    def adjoint_solver_complex_fields(
        self, design_params, frequencies=None, need_gradient=True
    ):
        matgrid = mp.MaterialGrid(
            mp.Vector3(self.Nx, self.Ny),
            mp.air,
            self.silicon,
            weights=np.ones((self.Nx, self.Ny)),
        )

        matgrid_region = mpa.DesignRegion(
            matgrid,
            volume=mp.Volume(
                center=mp.Vector3(),
                size=mp.Vector3(
                    self.design_region_size.x, self.design_region_size.y, 0
                ),
            ),
        )

        geometry = [
            mp.Block(
                center=matgrid_region.center, size=matgrid_region.size, material=matgrid
            )
        ]

        sim = mp.Simulation(
            resolution=self.resolution,
            cell_size=self.cell_size,
            default_material=self.silicon,
            k_point=self.k_point,
            boundary_layers=self.pml_x,
            sources=self.pt_source,
            geometry=geometry,
        )

        if not frequencies:
            frequencies = [self.fcen]

        obj_list = [
            mpa.FourierFields(
                sim, mp.Volume(center=mp.Vector3(0.9), size=mp.Vector3(0.2, 0.5)), mp.Dz
            )
        ]

        def J(dft_mon):
            return npa.power(npa.abs(dft_mon[:, 3, 9]), 2)

        opt = mpa.OptimizationProblem(
            simulation=sim,
            objective_functions=J,
            objective_arguments=obj_list,
            design_regions=[matgrid_region],
            frequencies=frequencies,
        )

        if need_gradient:
            f, dJ_du = opt([design_params])
            return f, dJ_du
        else:
            f = opt([design_params], need_gradient=False)
            return f[0]

    def adjoint_solver_damping(
        self, design_params, frequencies=None, mat2=None, need_gradient=True
    ):
        matgrid = mp.MaterialGrid(
            mp.Vector3(self.Nx, self.Ny),
            mp.air,
            self.silicon if mat2 is None else mat2,
            weights=np.ones((self.Nx, self.Ny)),
            damping=np.pi * self.fcen,
        )

        matgrid_region = mpa.DesignRegion(
            matgrid,
            volume=mp.Volume(
                center=mp.Vector3(),
                size=mp.Vector3(
                    self.design_region_size.x, self.design_region_size.y, 0
                ),
            ),
        )

        matgrid_geometry = [
            mp.Block(
                center=matgrid_region.center, size=matgrid_region.size, material=matgrid
            )
        ]

        geometry = self.waveguide_geometry + matgrid_geometry

        sim = mp.Simulation(
            resolution=self.resolution,
            cell_size=self.cell_size,
            boundary_layers=self.pml_xy,
            sources=self.mode_source,
            geometry=geometry,
        )

        if not frequencies:
            frequencies = [self.fcen]

        obj_list = [
            mpa.EigenmodeCoefficient(
                sim,
                mp.Volume(
                    center=mp.Vector3(0.5 * self.sxy - self.dpml),
                    size=mp.Vector3(0, self.sxy - 2 * self.dpml, 0),
                ),
                1,
                eig_parity=self.eig_parity,
            )
        ]

        def J(mode_mon):
            return npa.power(npa.abs(mode_mon), 2)

        opt = mpa.OptimizationProblem(
            simulation=sim,
            objective_functions=J,
            objective_arguments=obj_list,
            design_regions=[matgrid_region],
            frequencies=frequencies,
            minimum_run_time=150,
        )

        if need_gradient:
            f, dJ_du = opt([design_params])
            return f, dJ_du
        else:
            f = opt([design_params], need_gradient=False)
            return f[0]

    def adjoint_solver_two_objfunc(
        self,
        design_params,
        frequencies=None,
        need_gradient=True,
    ):
        # Compute the incident fields of the mode source
        # in the straight waveguide for use as normalization
        # of the reflectance (S11) measurement.
        ref_sim = mp.Simulation(
            resolution=self.resolution,
            cell_size=self.cell_size,
            boundary_layers=self.pml_xy,
            sources=self.mode_source,
            geometry=self.waveguide_geometry,
        )
        dft_mon = ref_sim.add_mode_monitor(
            frequencies,
            mp.ModeRegion(
                center=self.refl_pt,
                size=mp.Vector3(0, self.sxy - 2 * self.dpml, 0),
            ),
            yee_grid=True,
        )
        ref_sim.run(until_after_sources=20)
        subtracted_dft_fields = ref_sim.get_flux_data(dft_mon)
        input_flux = np.array(mp.get_fluxes(dft_mon))

        matgrid = mp.MaterialGrid(
            mp.Vector3(self.Nx, self.Ny),
            mp.air,
            self.silicon,
            weights=np.ones((self.Nx, self.Ny)),
        )

        matgrid_region = mpa.DesignRegion(
            matgrid,
            volume=mp.Volume(
                center=mp.Vector3(),
                size=mp.Vector3(
                    self.design_region_size.x, self.design_region_size.y, 0
                ),
            ),
        )

        matgrid_geometry = [
            mp.Block(
                center=matgrid_region.center,
                size=matgrid_region.size,
                material=matgrid,
            )
        ]

        geometry = self.waveguide_geometry + matgrid_geometry

        sim = mp.Simulation(
            resolution=self.resolution,
            cell_size=self.cell_size,
            boundary_layers=self.pml_xy,
            sources=self.mode_source,
            geometry=geometry,
        )

        obj_list = [
            mpa.EigenmodeCoefficient(
                sim,
                mp.Volume(
                    center=self.refl_pt,
                    size=mp.Vector3(0, self.sxy - 2 * self.dpml, 0),
                ),
                1,
                forward=False,
                subtracted_dft_fields=subtracted_dft_fields,
                eig_parity=self.eig_parity,
            ),
            mpa.EigenmodeCoefficient(
                sim,
                mp.Volume(
                    center=self.tran_pt,
                    size=mp.Vector3(0, self.sxy - 2 * self.dpml, 0),
                ),
                2,
                eig_parity=mp.ODD_Z,
            ),
        ]

        def J1(refl_mon, tran_mon):
            """Reflectance into first-order mode of Port 1."""
            return npa.power(npa.abs(refl_mon), 2) / input_flux

        def J2(refl_mon, tran_mon):
            """1-transmittance into second-order mode of Port 2."""
            return 1 - (npa.power(npa.abs(tran_mon), 2) / input_flux)

        opt = mpa.OptimizationProblem(
            simulation=sim,
            objective_functions=[J1, J2],
            objective_arguments=obj_list,
            design_regions=[matgrid_region],
            frequencies=frequencies,
        )

        if need_gradient:
            f0, dJ_du = opt([design_params])
            dJ_du_reflection = dJ_du[0]
            dJ_du_transmission = dJ_du[1]
            f0_reflection = f0[0]
            f0_transmission = f0[1]
            f0_merged = np.concatenate((f0_reflection, f0_transmission))
            nf = len(frequencies)
            grad = np.zeros((self.Nx * self.Ny, 2 * nf))
            if dJ_du_reflection.ndim < 2:
                dJ_du_reflection = np.expand_dims(dJ_du_reflection, axis=1)
            grad[:, :nf] = dJ_du_reflection
            if dJ_du_transmission.ndim < 2:
                dJ_du_transmission = np.expand_dims(dJ_du_transmission, axis=1)
            grad[:, nf:] = dJ_du_transmission
            return f0_merged, grad
        else:
            f0 = opt([design_params], need_gradient=False)
            f0_reflection = f0[0][0]
            f0_transmission = f0[0][1]
            f0_merged = np.concatenate((f0_reflection, f0_transmission))
            return f0_merged

    def mapping(self, x, filter_radius, eta, beta):
        filtered_field = mpa.conic_filter(
            x,
            filter_radius,
            self.design_region_size.x,
            self.design_region_size.y,
            self.design_region_resolution,
        )

        projected_field = mpa.tanh_projection(filtered_field, beta, eta)

        return projected_field.flatten()

    def test_DFT_fields(self):
        print("*** TESTING DFT OBJECTIVE ***")

        # test the single frequency and multi frequency cases
        for frequencies in [[self.fcen], [1 / 1.58, self.fcen, 1 / 1.53]]:
            # compute objective value and its gradient for unperturbed design
            unperturbed_val, unperturbed_grad = self.adjoint_solver(
                self.p, MonitorObject.DFT, frequencies
            )

            # compute objective value for perturbed design
            perturbed_val = self.adjoint_solver(
                self.p + self.dp, MonitorObject.DFT, frequencies, need_gradient=False
            )

            # compare directional derivative
            if unperturbed_grad.ndim < 2:
                unperturbed_grad = np.expand_dims(unperturbed_grad, axis=1)
            adj_dd = (self.dp[None, :] @ unperturbed_grad).flatten()
            fnd_dd = perturbed_val - unperturbed_val
            print(
                f"directional derivative:, {adj_dd} (adjoint solver), {fnd_dd} (finite difference)"
            )

            tol = 0.07 if mp.is_single_precision() else 0.006
            self.assertClose(adj_dd, fnd_dd, epsilon=tol)

    def test_eigenmode(self):
        print("*** TESTING EIGENMODE OBJECTIVE ***")

        # test the single frequency and multi frequency case
        for frequencies in [[self.fcen], [1 / 1.58, self.fcen, 1 / 1.53]]:
            # compute objective value and its gradient for unperturbed design
            unperturbed_val, unperturbed_grad = self.adjoint_solver(
                self.p, MonitorObject.EIGENMODE, frequencies
            )

            # compute objective for perturbed design
            perturbed_val = self.adjoint_solver(
                self.p + self.dp,
                MonitorObject.EIGENMODE,
                frequencies,
                need_gradient=False,
            )

            # compare directional derivative
            if unperturbed_grad.ndim < 2:
                unperturbed_grad = np.expand_dims(unperturbed_grad, axis=1)
            adj_dd = (self.dp[None, :] @ unperturbed_grad).flatten()
            fnd_dd = perturbed_val - unperturbed_val
            print(
                f"directional derivative:, {adj_dd} (adjoint solver), {fnd_dd} (finite difference)"
            )

            tol = 0.04 if mp.is_single_precision() else 0.01
            self.assertClose(adj_dd, fnd_dd, epsilon=tol)

    def test_ldos(self):
        print("*** TESTING LDOS OBJECTIVE ***")

        # test the single frequency and multi frequency cases
        for frequencies in [[self.fcen], [1 / 1.58, self.fcen, 1 / 1.53]]:
            # compute objective value and its gradient for unperturbed design
            unperturbed_val, unperturbed_grad = self.adjoint_solver(
                self.p, MonitorObject.LDOS, frequencies
            )

            # compute objective for perturbed design
            perturbed_val = self.adjoint_solver(
                self.p + self.dp, MonitorObject.LDOS, frequencies, need_gradient=False
            )

            # compare directional derivative
            if unperturbed_grad.ndim < 2:
                unperturbed_grad = np.expand_dims(unperturbed_grad, axis=1)
            adj_dd = (self.dp[None, :] @ unperturbed_grad).flatten()
            fnd_dd = perturbed_val - unperturbed_val
            print(
                f"directional derivative:, {adj_dd} (adjoint solver), {fnd_dd} (finite difference)"
            )

            tol = 0.07 if mp.is_single_precision() else 0.006
            self.assertClose(adj_dd, fnd_dd, epsilon=tol)

    def test_gradient_backpropagation(self):
        print("*** TESTING GRADIENT BACKPROPAGATION ***")

        # filter/thresholding parameters
        filter_radius = 0.21985
        eta = 0.49093
        beta = 4.0698

        for frequencies in [[self.fcen], [1 / 1.58, self.fcen, 1 / 1.53]]:
            mapped_p = self.mapping(self.p, filter_radius, eta, beta)

            # compute objective value and its gradient for unperturbed design
            unperturbed_val, unperturbed_grad = self.adjoint_solver(
                mapped_p, MonitorObject.EIGENMODE, frequencies
            )

            # backpropagate the gradient using vector-Jacobian product
            if len(frequencies) > 1:
                unperturbed_grad_backprop = np.zeros(unperturbed_grad.shape)
                for i in range(len(frequencies)):
                    unperturbed_grad_backprop[:, i] = tensor_jacobian_product(
                        self.mapping, 0
                    )(self.p, filter_radius, eta, beta, unperturbed_grad[:, i])
            else:
                unperturbed_grad_backprop = tensor_jacobian_product(self.mapping, 0)(
                    self.p, filter_radius, eta, beta, unperturbed_grad
                )

            # compute objective for perturbed design
            perturbed_val = self.adjoint_solver(
                self.mapping(self.p + self.dp, filter_radius, eta, beta),
                MonitorObject.EIGENMODE,
                frequencies,
                need_gradient=False,
            )

            # compare directional derivative
            if unperturbed_grad_backprop.ndim < 2:
                unperturbed_grad_backprop = np.expand_dims(
                    unperturbed_grad_backprop, axis=1
                )
            adj_dd = (self.dp[None, :] @ unperturbed_grad_backprop).flatten()
            fnd_dd = perturbed_val - unperturbed_val
            print(
                f"directional derivative:, {adj_dd} (adjoint solver), {fnd_dd} (finite difference)"
            )

            tol = 0.025 if mp.is_single_precision() else 0.01
            self.assertClose(adj_dd, fnd_dd, epsilon=tol)

    def test_complex_fields(self):
        print("*** TESTING COMPLEX FIELDS ***")

        for frequencies in [[self.fcen], [1 / 1.58, self.fcen, 1 / 1.53]]:
            # compute objective value and its gradient for unperturbed design
            unperturbed_val, unperturbed_grad = self.adjoint_solver_complex_fields(
                self.p, frequencies
            )

            # compute objective value perturbed design
            perturbed_val = self.adjoint_solver_complex_fields(
                self.p + self.dp, frequencies, need_gradient=False
            )

            # compare directional derivative
            if unperturbed_grad.ndim < 2:
                unperturbed_grad = np.expand_dims(unperturbed_grad, axis=1)
            adj_dd = (self.dp[None, :] @ unperturbed_grad).flatten()
            fnd_dd = perturbed_val - unperturbed_val
            print(
                f"directional derivative:, {adj_dd} (adjoint solver), {fnd_dd} (finite difference)"
            )

            tol = 0.06 if mp.is_single_precision() else 0.01
            self.assertClose(adj_dd, fnd_dd, epsilon=tol)

    def test_damping(self):
        print("*** TESTING CONDUCTIVITY ***")

        for frequencies in [[1 / 1.58, self.fcen, 1 / 1.53]]:
            # compute objective value and its gradient for unperturbed design
            unperturbed_val, unperturbed_grad = self.adjoint_solver_damping(
                self.p, frequencies
            )

            # compute objective value perturbed design
            perturbed_val = self.adjoint_solver_damping(
                self.p + self.dp, frequencies, need_gradient=False
            )

            # compare directional derivative
            if unperturbed_grad.ndim < 2:
                unperturbed_grad = np.expand_dims(unperturbed_grad, axis=1)
            adj_dd = (self.dp[None, :] @ unperturbed_grad).flatten()
            fnd_dd = perturbed_val - unperturbed_val
            print(
                f"directional derivative:, {adj_dd} (adjoint solver), {fnd_dd} (finite difference)"
            )

            tol = 0.06 if mp.is_single_precision() else 0.04
            self.assertClose(adj_dd, fnd_dd, epsilon=tol)

    def test_offdiagonal(self):
        print("*** TESTING ANISOTROPIC ε ***")
        filt = lambda x: mpa.conic_filter(
            x.reshape((self.Nx, self.Ny)),
            0.25,
            self.design_region_size.x,
            self.design_region_size.y,
            self.design_region_resolution,
        ).flatten()

        # test the single frequency and multi frequency case
        for frequencies in [[self.fcen], [1 / 1.58, self.fcen, 1 / 1.53]]:
            # compute objective value and its gradient for unperturbed design
            unperturbed_val, unperturbed_grad = self.adjoint_solver(
                filt(self.p), MonitorObject.EIGENMODE, frequencies, self.sapphire
            )

            # backpropagate the gradient using vector-Jacobian product
            if len(frequencies) > 1:
                unperturbed_grad_backprop = np.zeros(unperturbed_grad.shape)
                for i in range(len(frequencies)):
                    unperturbed_grad_backprop[:, i] = tensor_jacobian_product(filt, 0)(
                        self.p, unperturbed_grad[:, i]
                    )
            else:
                unperturbed_grad_backprop = tensor_jacobian_product(filt, 0)(
                    self.p, unperturbed_grad
                )

            # compute objective value perturbed design
            perturbed_val = self.adjoint_solver(
                filt(self.p + self.dp),
                MonitorObject.EIGENMODE,
                frequencies,
                self.sapphire,
                need_gradient=False,
            )

            # compare directional derivative
            if unperturbed_grad_backprop.ndim < 2:
                unperturbed_grad_backprop = np.expand_dims(
                    unperturbed_grad_backprop, axis=1
                )
            adj_dd = (self.dp[None, :] @ unperturbed_grad_backprop).flatten()
            fnd_dd = perturbed_val - unperturbed_val
            print(
                f"directional derivative:, {adj_dd} (adjoint solver), {fnd_dd} (finite difference)"
            )

            tol = 0.1 if mp.is_single_precision() else 0.04
            self.assertClose(adj_dd, fnd_dd, epsilon=tol)

    def test_two_objfunc(self):
        print("*** TESTING TWO OBJECTIVE FUNCTIONS***")

        # test the single frequency and multi frequency case
        for frequencies in [[self.fcen], [1 / 1.58, self.fcen, 1 / 1.53]]:
            # compute objective value and its gradient for unperturbed design
            unperturbed_val, unperturbed_grad = self.adjoint_solver_two_objfunc(
                self.p, frequencies
            )

            # compute objective value for perturbed design
            perturbed_val = self.adjoint_solver_two_objfunc(
                self.p + self.dp,
                frequencies,
                need_gradient=False,
            )

            nfrq = len(frequencies)
            tol = 0.05 if mp.is_single_precision() else 0.01
            for m in [0, 1]:
                frq_slice = slice(0, nfrq, 1) if m == 0 else slice(nfrq, 2 * nfrq, 1)
                adj_dd = (self.dp[None, :] @ unperturbed_grad[:, frq_slice]).flatten()
                fnd_dd = perturbed_val[frq_slice] - unperturbed_val[frq_slice]
                print(
                    f"directional derivative:, "
                    f"{adj_dd} (adjoint solver), "
                    f"{fnd_dd} (finite difference)"
                )
                self.assertClose(adj_dd, fnd_dd, epsilon=tol)


if __name__ == "__main__":
    unittest.main()



## linear-chirped pulse planewave with higher frequencies at the front (down-chirp)
import numpy as np

import meep as mp

resolution = 40

dpml = 2
pml_layers = [mp.PML(thickness=dpml, direction=mp.X)]

sx = 40
sy = 6
cell_size = mp.Vector3(sx + 2 * dpml, sy)

v0 = 1.0  # pulse center frequency
a = 0.2  # Gaussian envelope half-width
b = -0.5  # linear chirp rate (positive: up-chirp, negative: down-chirp)
t0 = 15  # peak time

chirp = lambda t: np.exp(1j * 2 * np.pi * v0 * (t - t0)) * np.exp(
    -a * (t - t0) ** 2 + 1j * b * (t - t0) ** 2
)

sources = [
    mp.Source(
        src=mp.CustomSource(src_func=chirp),
        center=mp.Vector3(-0.5 * sx),
        size=mp.Vector3(y=sy),
        component=mp.Ez,
    )
]

sim = mp.Simulation(
    cell_size=cell_size,
    boundary_layers=pml_layers,
    resolution=resolution,
    k_point=mp.Vector3(),
    sources=sources,
    symmetries=[mp.Mirror(mp.Y)],
)

sim.run(
    mp.in_volume(
        mp.Volume(center=mp.Vector3(), size=mp.Vector3(sx, sy)),
        mp.at_every(2.7, mp.output_efield_z),
    ),
    until=t0 + 50,
)



import unittest

import meep.adjoint as mpa
import numpy as np
import parameterized
from meep.materials import Co, SiN

import meep as mp


class TestGetEpsilonGrid(unittest.TestCase):
    def setUp(self):
        resolution = 60
        self.cell_size = mp.Vector3(1.0, 1.0, 0)

        matgrid_resolution = 200
        matgrid_size = mp.Vector3(1.0, 1.0, mp.inf)
        Nx, Ny = int(matgrid_size.x * matgrid_resolution), int(
            matgrid_size.y * matgrid_resolution
        )
        x = np.linspace(-0.5 * matgrid_size.x, 0.5 * matgrid_size.x, Nx)
        y = np.linspace(-0.5 * matgrid_size.y, 0.5 * matgrid_size.y, Ny)
        xv, yv = np.meshgrid(x, y)
        rad = 0.201943
        w = 0.104283
        weights = np.logical_and(
            np.sqrt(np.square(xv) + np.square(yv)) > rad,
            np.sqrt(np.square(xv) + np.square(yv)) < rad + w,
            dtype=np.double,
        )

        matgrid = mp.MaterialGrid(
            mp.Vector3(Nx, Ny),
            mp.air,
            mp.Medium(index=3.5),
            weights=weights,
            do_averaging=False,
            beta=0,
            eta=0.5,
        )

        geometry = [
            mp.Cylinder(
                center=mp.Vector3(0.35, 0.1),
                radius=0.1,
                height=mp.inf,
                material=mp.Medium(index=1.5),
            ),
            mp.Block(
                center=mp.Vector3(-0.15, -0.2),
                size=mp.Vector3(0.2, 0.24, mp.inf),
                material=SiN,
            ),
            mp.Block(
                center=mp.Vector3(-0.2, 0.2),
                size=mp.Vector3(0.4, 0.4, mp.inf),
                material=matgrid,
            ),
            mp.Prism(
                vertices=[
                    mp.Vector3(0.05, 0.45),
                    mp.Vector3(0.32, 0.22),
                    mp.Vector3(0.15, 0.10),
                ],
                height=0.5,
                material=Co,
            ),
        ]

        self.sim = mp.Simulation(
            resolution=resolution,
            cell_size=self.cell_size,
            geometry=geometry,
            eps_averaging=False,
        )
        self.sim.init_sim()

    @parameterized.parameterized.expand(
        [
            (mp.Vector3(0.2, 0.2), 1.1),
            (mp.Vector3(-0.2, 0.1), 0.7),
            (mp.Vector3(-0.2, -0.25), 0.55),
            (mp.Vector3(0.4, 0.1), 0),
        ]
    )
    def test_get_epsilon_grid(self, pt, freq):
        eps_grid = self.sim.get_epsilon_grid(
            np.array([pt.x]), np.array([pt.y]), np.array([0]), freq
        )
        eps_pt = self.sim.get_epsilon_point(pt, freq)
        print(f"eps:, ({pt.x},{pt.y}), {eps_grid}, {eps_pt}")
        self.assertAlmostEqual(np.real(eps_grid), np.real(eps_pt), places=6)
        self.assertAlmostEqual(np.imag(eps_grid), np.imag(eps_pt), places=6)


if __name__ == "__main__":
    unittest.main()



import unittest

from meep.materials import Ag, Cr, Ge, InP, LiNbO3, Si, SiO2_aniso


class TestMaterialsLibrary(unittest.TestCase):
    def test_materials_library(self):
        self.assertAlmostEqual(InP.epsilon(1 / 3.3)[0][0], (3.1031) ** 2, places=2)

        self.assertAlmostEqual(Ge.epsilon(1 / 6.8)[0][0], (4.0091) ** 2, places=2)

        self.assertAlmostEqual(Si.epsilon(1 / 1.55)[0][0], (3.4777) ** 2, places=2)

        self.assertAlmostEqual(LiNbO3.epsilon(1 / 1.55)[0][0], (2.2111) ** 2, places=2)
        self.assertAlmostEqual(LiNbO3.epsilon(1 / 1.55)[1][1], (2.2111) ** 2, places=2)
        self.assertAlmostEqual(LiNbO3.epsilon(1 / 1.55)[2][2], (2.1376) ** 2, places=2)

        self.assertAlmostEqual(
            SiO2_aniso.epsilon(1 / 1.55)[0][0], (1.5277) ** 2, places=2
        )
        self.assertEqual(SiO2_aniso.epsilon(1 / 1.55)[1][0], 0)
        self.assertAlmostEqual(
            SiO2_aniso.epsilon(1 / 1.55)[1][1], (1.5277) ** 2, places=2
        )
        self.assertAlmostEqual(
            SiO2_aniso.epsilon(1 / 1.55)[2][2], (1.5362) ** 2, places=2
        )

        self.assertAlmostEqual(
            Ag.epsilon(1 / 0.65)[0][0], (0.14623 + 1j * 3.9367) ** 2, places=2
        )

        self.assertAlmostEqual(
            Cr.epsilon(1 / 0.71)[0][0], (3.8275 + 1j * 4.3457) ** 2, places=2
        )

        try:
            Ag.epsilon(1 / 0.2)[0][0]
        except ValueError:
            pass
        else:
            raise AssertionError("Ag is not defined at a wavelength of 0.2 μm")


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np
from utils import ApproxComparisonTestCase

import meep as mp


class TestArrayMetadata(ApproxComparisonTestCase):
    def test_array_metadata(self):
        resolution = 25

        n = 3.4
        w = 1
        r = 1
        pad = 4
        dpml = 2

        sxy = 2 * (r + w + pad + dpml)
        cell_size = mp.Vector3(sxy, sxy)

        nonpml_vol = mp.Volume(
            mp.Vector3(), size=mp.Vector3(sxy - 2 * dpml, sxy - 2 * dpml)
        )

        geometry = [
            mp.Cylinder(radius=r + w, material=mp.Medium(index=n)),
            mp.Cylinder(radius=r),
        ]

        fcen = 0.118
        df = 0.08

        symmetries = [mp.Mirror(mp.X, phase=-1), mp.Mirror(mp.Y, phase=+1)]

        pml_layers = [mp.PML(dpml)]

        # CW source
        src = [
            mp.Source(mp.ContinuousSource(fcen, fwidth=df), mp.Ez, mp.Vector3(r + 0.1)),
            mp.Source(
                mp.ContinuousSource(fcen, fwidth=df),
                mp.Ez,
                mp.Vector3(-(r + 0.1)),
                amplitude=-1,
            ),
        ]

        sim = mp.Simulation(
            cell_size=cell_size,
            geometry=geometry,
            sources=src,
            resolution=resolution,
            force_complex_fields=True,
            symmetries=symmetries,
            boundary_layers=pml_layers,
        )

        sim.init_sim()
        sim.solve_cw(1e-5 if mp.is_single_precision() else 1e-6, 1000, 10)

        def electric_energy(r, ez, eps):
            return np.real(eps * np.conj(ez) * ez)

        def vec_func(r):
            return r.x**2 + 2 * r.y**2

        electric_energy_total = sim.integrate_field_function(
            [mp.Ez, mp.Dielectric], electric_energy, nonpml_vol
        )
        electric_energy_max = sim.max_abs_field_function(
            [mp.Ez, mp.Dielectric], electric_energy, nonpml_vol
        )
        vec_func_total = sim.integrate_field_function([], vec_func, nonpml_vol)
        cw_modal_volume = (electric_energy_total / electric_energy_max) * vec_func_total

        sim.reset_meep()

        # pulsed source
        src = [
            mp.Source(mp.GaussianSource(fcen, fwidth=df), mp.Ez, mp.Vector3(r + 0.1)),
            mp.Source(
                mp.GaussianSource(fcen, fwidth=df),
                mp.Ez,
                mp.Vector3(-(r + 0.1)),
                amplitude=-1,
            ),
        ]

        sim = mp.Simulation(
            cell_size=cell_size,
            geometry=geometry,
            k_point=mp.Vector3(),
            sources=src,
            resolution=resolution,
            symmetries=symmetries,
            boundary_layers=pml_layers,
        )

        dft_obj = sim.add_dft_fields([mp.Ez], fcen, 0, 1, where=nonpml_vol)
        sim.run(until_after_sources=100)

        Ez = sim.get_dft_array(dft_obj, mp.Ez, 0)
        (X, Y, Z, W) = sim.get_array_metadata(dft_cell=dft_obj)
        Eps = sim.get_array(vol=nonpml_vol, component=mp.Dielectric)
        EpsE2 = np.real(Eps * np.conj(Ez) * Ez)
        xm, ym = np.meshgrid(X, Y)
        vec_func_sum = np.sum(W * (xm**2 + 2 * ym**2))
        pulse_modal_volume = np.sum(W * EpsE2) / np.max(EpsE2) * vec_func_sum

        tol = 5e-2 if mp.is_single_precision() else 1e-2
        self.assertClose(cw_modal_volume / pulse_modal_volume, 1.0, epsilon=tol)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
from distutils.core import setup

setup(
    name="python-markdown-math",
    description="Math extension for Python-Markdown",
    author="Dmitry Shachnev",
    author_email="mitya57@gmail.com",
    version="0.1",
    url="https://github.com/mitya57/python-markdown-math",
    py_modules=["mdx_math"],
    license="BSD",
)



import importlib

import gdspy
import numpy as np
from matplotlib import pyplot as plt

import meep as mp

# core and cladding materials
Si = mp.Medium(index=3.4)
SiO2 = mp.Medium(index=1.4)

# layer numbers for GDS file
RING_LAYER = 0
SOURCE0_LAYER = 1
SOURCE1_LAYER = 2
MONITOR_LAYER = 3
SIMULATION_LAYER = 4

resolution = 50  # pixels/μm
dpml = 1  # thickness of PML
zmin = 0  # minimum z value of simulation domain (0 for 2D)
zmax = 0  # maximum z value of simulation domain (0 for 2D)


def create_ring_gds(radius, width):
    # Reload the library each time to prevent gds library name clashes
    importlib.reload(gdspy)

    ringCell = gdspy.Cell(f"ring_resonator_r{radius}_w{width}")

    # Draw the ring
    ringCell.add(
        gdspy.Round(
            (0, 0),
            inner_radius=radius - width / 2,
            radius=radius + width / 2,
            layer=RING_LAYER,
        )
    )

    # Draw the first source
    ringCell.add(
        gdspy.Rectangle((radius - width, 0), (radius + width, 0), SOURCE0_LAYER)
    )

    # Draw the second source
    ringCell.add(
        gdspy.Rectangle((-radius - width, 0), (-radius + width, 0), SOURCE1_LAYER)
    )

    # Draw the monitor location
    ringCell.add(
        gdspy.Rectangle((radius - width / 2, 0), (radius + width / 2, 0), MONITOR_LAYER)
    )

    # Draw the simulation domain
    pad = 2  # padding between waveguide and edge of PML
    ringCell.add(
        gdspy.Rectangle(
            (-radius - width / 2 - pad, -radius - width / 2 - pad),
            (radius + width / 2 + pad, radius + width / 2 + pad),
            SIMULATION_LAYER,
        )
    )

    filename = f"ring_r{radius}_w{width}.gds"
    gdspy.write_gds(filename, unit=1.0e-6, precision=1.0e-9)

    return filename


def find_modes(filename, wvl=1.55, bw=0.05):
    # Read in the ring structure
    geometry = mp.get_GDSII_prisms(Si, filename, RING_LAYER, -100, 100)

    cell = mp.GDSII_vol(filename, SIMULATION_LAYER, zmin, zmax)

    src_vol0 = mp.GDSII_vol(filename, SOURCE0_LAYER, zmin, zmax)
    src_vol1 = mp.GDSII_vol(filename, SOURCE1_LAYER, zmin, zmax)

    mon_vol = mp.GDSII_vol(filename, MONITOR_LAYER, zmin, zmax)

    fcen = 1 / wvl
    df = bw * fcen

    src = [
        mp.Source(mp.GaussianSource(fcen, fwidth=df), component=mp.Hz, volume=src_vol0),
        mp.Source(
            mp.GaussianSource(fcen, fwidth=df),
            component=mp.Hz,
            volume=src_vol1,
            amplitude=-1,
        ),
    ]

    sim = mp.Simulation(
        cell_size=cell.size,
        geometry=geometry,
        sources=src,
        resolution=resolution,
        boundary_layers=[mp.PML(dpml)],
        default_material=SiO2,
    )

    h = mp.Harminv(mp.Hz, mon_vol.center, fcen, df)

    sim.run(mp.after_sources(h), until_after_sources=100)

    plt.figure()
    sim.plot2D(fields=mp.Hz, eps_parameters={"contour": True})
    plt.savefig("ring_fields.png", bbox_inches="tight", dpi=150)

    wvl = np.array([1 / m.freq for m in h.modes])
    Q = np.array([m.Q for m in h.modes])

    sim.reset_meep()

    return wvl, Q


if __name__ == "__main__":
    filename = create_ring_gds(2.0, 0.5)
    wvls, Qs = find_modes(filename, 1.55, 0.05)
    for w, Q in zip(wvls, Qs):
        print(f"mode: {w}, {Q}")



import unittest

import numpy as np

import meep as mp


class TestMaterialDispersion(unittest.TestCase):
    def test_material_dispersion_with_user_material(self):
        susceptibilities = [
            mp.LorentzianSusceptibility(frequency=1.1, gamma=1e-5, sigma=0.5),
            mp.LorentzianSusceptibility(frequency=0.5, gamma=0.1, sigma=2e-5),
        ]

        def mat_func(p):
            return mp.Medium(epsilon=2.25, E_susceptibilities=susceptibilities)

        fcen = 1.0
        df = 2.0

        sources = mp.Source(
            mp.GaussianSource(fcen, fwidth=df), component=mp.Ez, center=mp.Vector3()
        )

        kmin = 0.3
        kmax = 2.2
        k_interp = 5

        kpts = mp.interpolate(k_interp, [mp.Vector3(kmin), mp.Vector3(kmax)])

        self.sim = mp.Simulation(
            cell_size=mp.Vector3(),
            geometry=[],
            sources=[sources],
            material_function=mat_func,
            default_material=mp.air,
            resolution=20,
        )

        all_freqs = self.sim.run_k_points(200, kpts)
        res = [f.real for fs in all_freqs for f in fs]

        expected = [
            0.1999342026399106,
            0.41053963810375294,
            0.6202409070451909,
            0.8285737385146619,
            1.0350739448523063,
            1.2392775309110078,
            1.4407208712852109,
        ]

        np.testing.assert_allclose(expected, res)


if __name__ == "__main__":
    unittest.main()



## moving point charge with superluminal phase velocity in dielectric media emitting Cherenkov radiation
import meep as mp

sx = 60
sy = 60
cell_size = mp.Vector3(sx, sy, 0)

dpml = 1.0
pml_layers = [mp.PML(thickness=dpml)]

v = 0.7  # velocity of point charge

symmetries = [mp.Mirror(direction=mp.Y)]

sim = mp.Simulation(
    resolution=10,
    cell_size=cell_size,
    default_material=mp.Medium(index=1.5),
    symmetries=symmetries,
    boundary_layers=pml_layers,
)


def move_source(sim):
    sim.change_sources(
        [
            mp.Source(
                mp.ContinuousSource(frequency=1e-10),
                component=mp.Ex,
                center=mp.Vector3(-0.5 * sx + dpml + v * sim.meep_time()),
            )
        ]
    )


sim.run(
    move_source,
    mp.at_every(2, mp.output_png(mp.Hz, "-vZc dkbluered -M 1")),
    until=sx / v,
)



import matplotlib
import numpy as np

import meep as mp

matplotlib.use("agg")
import matplotlib.pyplot as plt

resolution = 20  # pixels/μm

fcen = 0.25  # pulse center frequency
df = 0.2  # pulse width (in frequency)

eps = 13  # dielectric constant of waveguide
w = 1.2  # width of waveguide
r = 0.36  # radius of holes
d = 1.4  # defect spacing (ordinary spacing = 1)
N = 3  # number of holes on either side of defect

dpad = 32  # padding between last hole and PML edge
dpml = 0.5 / (fcen - 0.5 * df)  # PML thickness (> half the largest wavelength)
sx = 2 * (dpad + dpml + N) + d - 1  # size of cell in x direction

d1 = 0.2  # y-distance from waveguide edge to near2far surface
d2 = 2.0  # y-distance from near2far surface to far-field line
sy = w + 2 * (d1 + d2 + dpml)  # size of cell in y direction (perpendicular to wvg.)

cell = mp.Vector3(sx, sy, 0)

geometry = [
    mp.Block(
        center=mp.Vector3(),
        size=mp.Vector3(mp.inf, w, mp.inf),
        material=mp.Medium(epsilon=eps),
    )
]

for i in range(N):
    geometry.append(mp.Cylinder(r, center=mp.Vector3(d / 2 + i)))
    geometry.append(mp.Cylinder(r, center=mp.Vector3(d / -2 - i)))

pml_layers = [mp.PML(dpml)]

sources = [
    mp.Source(
        src=mp.GaussianSource(fcen, fwidth=df), component=mp.Hz, center=mp.Vector3()
    )
]

symmetries = [mp.Mirror(mp.X, phase=-1), mp.Mirror(mp.Y, phase=-1)]

sim = mp.Simulation(
    cell_size=cell,
    geometry=geometry,
    sources=sources,
    symmetries=symmetries,
    boundary_layers=pml_layers,
    resolution=resolution,
)

nearfield = sim.add_near2far(
    fcen,
    0,
    1,
    mp.Near2FarRegion(mp.Vector3(0, 0.5 * w + d1), size=mp.Vector3(sx - 2 * dpml)),
    mp.Near2FarRegion(
        mp.Vector3(-0.5 * sx + dpml, 0.5 * w + 0.5 * d1),
        size=mp.Vector3(0, d1),
        weight=-1.0,
    ),
    mp.Near2FarRegion(
        mp.Vector3(0.5 * sx - dpml, 0.5 * w + 0.5 * d1), size=mp.Vector3(0, d1)
    ),
)

mon = sim.add_dft_fields(
    [mp.Hz],
    fcen,
    0,
    1,
    center=mp.Vector3(0, 0.5 * w + d1 + d2),
    size=mp.Vector3(sx - 2 * (dpad + dpml), 0),
)

sim.run(until_after_sources=mp.stop_when_dft_decayed())

sim.plot2D()
if mp.am_master():
    plt.savefig(
        f"cavity_farfield_plot2D_dpad{dpad}_{d1}_{d2}.png", bbox_inches="tight", dpi=150
    )

Hz_mon = sim.get_dft_array(mon, mp.Hz, 0)

(x, y, z, w) = sim.get_array_metadata(dft_cell=mon)

ff = []
for xc in x:
    ff_pt = sim.get_farfield(nearfield, mp.Vector3(xc, y[0]))
    ff.append(ff_pt[5])
ff = np.array(ff)

if mp.am_master():
    plt.figure()
    plt.subplot(1, 3, 1)
    plt.plot(np.real(Hz_mon), "bo-", label="DFT")
    plt.plot(np.real(ff), "ro-", label="N2F")
    plt.legend()
    plt.xlabel("$x$ (μm)")
    plt.ylabel("real(Hz)")

    plt.subplot(1, 3, 2)
    plt.plot(np.imag(Hz_mon), "bo-", label="DFT")
    plt.plot(np.imag(ff), "ro-", label="N2F")
    plt.legend()
    plt.xlabel("$x$ (μm)")
    plt.ylabel("imag(Hz)")

    plt.subplot(1, 3, 3)
    plt.plot(np.abs(Hz_mon), "bo-", label="DFT")
    plt.plot(np.abs(ff), "ro-", label="N2F")
    plt.legend()
    plt.xlabel("$x$ (μm)")
    plt.ylabel("|Hz|")

    plt.suptitle(
        f"comparison of near2far and actual DFT fields\n dpad={dpad}, d1={d1}, d2={d2}"
    )
    plt.subplots_adjust(wspace=0.6)
    plt.savefig(
        f"test_Hz_dft_vs_n2f_res{resolution}_dpad{dpad}_d1{d1}_d2{d2}.png",
        bbox_inches="tight",
        dpi=150,
    )



"""Handling of objective functions and objective quantities."""
import abc
from collections import namedtuple
from typing import Callable, List, Optional

import numpy as np
from meep.simulation import py_v3_to_vec, FluxData, NearToFarData

import meep as mp

from .filter_source import FilteredSource

Grid = namedtuple("Grid", ["x", "y", "z", "w"])


class ObjectiveQuantity(abc.ABC):
    """A differentiable objective quantity.

    Attributes:
        sim: the Meep simulation object with which the objective quantity is registered.
        frequencies: the frequencies at which the objective quantity is evaluated.
        num_freq: the number of frequencies at which the objective quantity is evaluated.
    """

    def __init__(self, sim):
        self.sim = sim
        self._eval = None
        self._frequencies = None

    @property
    def frequencies(self):
        return self._frequencies

    @property
    def num_freq(self):
        return len(self.frequencies)

    @abc.abstractmethod
    def __call__(self):
        """Evaluates the objective quantity."""

    @abc.abstractmethod
    def register_monitors(self, frequencies):
        """Registers monitors in the forward simulation."""

    @abc.abstractmethod
    def place_adjoint_source(self, dJ):
        """Places appropriate sources for the adjoint simulation."""

    def get_evaluation(self):
        """Evaluates the objective quantity."""
        if self._eval is not None:
            return self._eval
        else:
            raise RuntimeError(
                "You must first run a forward simulation before requesting the evaluation of an objective quantity."
            )

    def _adj_src_scale(self, include_resolution=True):
        """Calculates the scale for the adjoint sources."""
        T = self.sim.meep_time()
        dt = self.sim.fields.dt
        src = self._create_time_profile()

        if include_resolution:
            num_dims = self.sim._infer_dimensions(self.sim.k_point)
            dV = 1 / self.sim.resolution**num_dims
        else:
            dV = 1

        iomega = (1.0 - np.exp(-1j * (2 * np.pi * self._frequencies) * dt)) * (
            1.0 / dt
        )  # scaled frequency factor with discrete time derivative fix

        # an ugly way to calcuate the scaled dtft of the forward source
        y = np.array(
            [src.swigobj.current(t, dt) for t in np.arange(0, T, dt)]
        )  # time domain signal
        fwd_dtft = (
            np.matmul(
                np.exp(
                    1j
                    * 2
                    * np.pi
                    * self._frequencies[:, np.newaxis]
                    * np.arange(y.size)
                    * dt
                ),
                y,
            )
            * dt
            / np.sqrt(2 * np.pi)
        )  # dtft

        # Interestingly, the real parts of the DTFT and fourier transform match, but the imaginary parts are very different...
        # fwd_dtft = src.fourier_transform(src.frequency)
        """
        For some reason, there seems to be an additional phase
        factor at the center frequency that needs to be applied
        to *all* frequencies...
        """
        src_center_dtft = (
            np.matmul(
                np.exp(
                    1j
                    * 2
                    * np.pi
                    * np.array([src.frequency])[:, np.newaxis]
                    * np.arange(y.size)
                    * dt
                ),
                y,
            )
            * dt
            / np.sqrt(2 * np.pi)
        )
        adj_src_phase = np.exp(1j * np.angle(src_center_dtft)) * self.fwidth_scale

        if self._frequencies.size == 1:
            # Single frequency simulations. We need to drive it with a time profile.
            scale = dV * iomega / fwd_dtft / adj_src_phase  # final scale factor
        else:
            # multi frequency simulations
            scale = dV * iomega / adj_src_phase
        # compensate for the fact that real fields take the real part of the current,
        # which halves the Fourier amplitude at the positive frequency (Re[J] = (J + J*)/2)
        if self.sim.using_real_fields():
            scale *= 2
        return scale

    def _create_time_profile(self, fwidth_frac=0.1, adj_cutoff=5):
        """Creates a time domain waveform for normalizing the adjoint source(s).

        For single frequency objective functions, we should generate a guassian pulse with a reasonable
        bandwidth centered at said frequency.

        TODO:
        The user may specify a scalar valued objective function across multiple frequencies (e.g. MSE) in
        which case we should check that all the frequencies fit in the specified bandwidth.
        """
        self.fwidth_scale = np.exp(-2j * np.pi * adj_cutoff / fwidth_frac)
        return mp.GaussianSource(
            np.mean(self._frequencies),
            fwidth=fwidth_frac * np.mean(self._frequencies),
            cutoff=adj_cutoff,
        )


class EigenmodeCoefficient(ObjectiveQuantity):
    """A frequency-dependent eigenmode coefficient.
    Attributes:
        volume: the volume over which the eigenmode coefficient is calculated.
        mode: the eigenmode number.
        forward: whether the forward or backward mode coefficient is returned as
          the result of the evaluation.
        kpoint_func: an optional k-point function to use when evaluating the eigenmode
          coefficient. When specified, this overrides the effect of `forward`.
        kpoint_func_overlap_idx: the index of the mode coefficient to return when
          specifying `kpoint_func`. When specified, this overrides the effect of
          `forward` and should have a value of either 0 or 1.
        subtracted_dft_fields: the DFT fields obtained using `get_flux_data` from
          a previous normalization run. This is subtracted from the DFT fields
          of this mode monitor in order to improve the accuracy of the
          reflectance measurement (i.e., the $S_{11}$ scattering parameter).
          Default is None.
    """

    def __init__(
        self,
        sim: mp.Simulation,
        volume: mp.Volume,
        mode: int,
        forward: Optional[bool] = True,
        kpoint_func: Optional[Callable] = None,
        kpoint_func_overlap_idx: Optional[int] = 0,
        decimation_factor: Optional[int] = 0,
        subtracted_dft_fields: Optional[FluxData] = None,
        **kwargs
    ):
        """
        + **`sim` [ `Simulation` ]** —
        """
        super().__init__(sim)
        if kpoint_func_overlap_idx not in [0, 1]:
            raise ValueError(
                "`kpoint_func_overlap_idx` should be either 0 or 1, but got %d"
                % (kpoint_func_overlap_idx,)
            )
        self.volume = volume
        self.mode = mode
        self.forward = forward
        self.kpoint_func = kpoint_func
        self.kpoint_func_overlap_idx = kpoint_func_overlap_idx
        self.eigenmode_kwargs = kwargs
        self._monitor = None
        self._cscale = None
        self.decimation_factor = decimation_factor
        self.subtracted_dft_fields = subtracted_dft_fields

    def register_monitors(self, frequencies):
        self._frequencies = np.asarray(frequencies)
        self._monitor = self.sim.add_mode_monitor(
            frequencies,
            mp.ModeRegion(center=self.volume.center, size=self.volume.size),
            yee_grid=True,
            decimation_factor=self.decimation_factor,
        )
        if self.subtracted_dft_fields is not None:
            self.sim.load_minus_flux_data(
                self._monitor,
                self.subtracted_dft_fields,
            )
        return self._monitor

    def place_adjoint_source(self, dJ):
        dJ = np.atleast_1d(dJ)
        if dJ.ndim == 2:
            dJ = np.sum(dJ, axis=1)
        time_src = self._create_time_profile()
        da_dE = 0.5 * self._cscale
        scale = self._adj_src_scale()

        if self.kpoint_func:
            eig_kpoint = -1 * self.kpoint_func(time_src.frequency, self.mode)
        else:
            center_frequency = 0.5 * (
                np.min(self.frequencies) + np.max(self.frequencies)
            )
            direction = mp.Vector3(
                *(np.eye(3)[self._monitor.normal_direction] * np.abs(center_frequency))
            )
            eig_kpoint = -1 * direction if self.forward else direction

        if self._frequencies.size == 1:
            amp = da_dE * dJ * scale
            src = time_src
        else:
            scale = da_dE * dJ * scale
            src = FilteredSource(
                time_src.frequency,
                self._frequencies,
                scale,
                self.sim.fields.dt,
            )
            amp = 1
        source = mp.EigenModeSource(
            src,
            eig_band=self.mode,
            direction=mp.NO_DIRECTION,
            eig_kpoint=eig_kpoint,
            amplitude=amp,
            eig_match_freq=True,
            size=self.volume.size,
            center=self.volume.center,
            **self.eigenmode_kwargs,
        )
        return [source]

    def __call__(self):
        if self.kpoint_func:
            kpoint_func = self.kpoint_func
            overlap_idx = self.kpoint_func_overlap_idx
        else:
            center_frequency = 0.5 * (
                np.min(self.frequencies) + np.max(self.frequencies)
            )
            kpoint = mp.Vector3(
                *(np.eye(3)[self._monitor.normal_direction] * np.abs(center_frequency))
            )
            kpoint_func = lambda *not_used: kpoint if self.forward else -1 * kpoint
            overlap_idx = 0
        ob = self.sim.get_eigenmode_coefficients(
            self._monitor,
            [self.mode],
            direction=mp.NO_DIRECTION,
            kpoint_func=kpoint_func,
            **self.eigenmode_kwargs,
        )
        overlaps = ob.alpha.squeeze(axis=0)
        assert overlaps.ndim == 2
        self._eval = overlaps[:, overlap_idx]
        self._cscale = ob.cscale
        return self._eval


class FourierFields(ObjectiveQuantity):
    def __init__(
        self,
        sim: mp.Simulation,
        volume: mp.Volume,
        component: List[int],
        yee_grid: Optional[bool] = False,
        decimation_factor: Optional[int] = 0,
        subtracted_dft_fields: Optional[FluxData] = None,
    ):
        """ """
        super().__init__(sim)
        self.volume = sim._fit_volume_to_simulation(volume)
        self.component = component
        self.yee_grid = yee_grid
        self.decimation_factor = decimation_factor
        self.subtracted_dft_fields = subtracted_dft_fields

    def register_monitors(self, frequencies):
        self._frequencies = np.asarray(frequencies)
        self._monitor = self.sim.add_dft_fields(
            [self.component],
            self._frequencies,
            where=self.volume,
            yee_grid=self.yee_grid,
            decimation_factor=self.decimation_factor,
        )
        if self.subtracted_dft_fields is not None:
            self.sim.load_minus_flux_data(
                self._monitor,
                self.subtracted_dft_fields,
            )
        return self._monitor

    def place_adjoint_source(self, dJ):
        time_src = self._create_time_profile()
        sources = []

        mon_size = self.sim.fields.dft_monitor_size(
            self._monitor.swigobj, self.volume.swigobj, self.component
        )
        dJ = dJ.astype(np.complex128)
        if (
            np.prod(mon_size) * self.num_freq != dJ.size
            and np.prod(mon_size) * self.num_freq**2 != dJ.size
        ):
            raise ValueError("The format of J is incorrect!")

        # The objective function J is a vector. Each component corresponds to a frequency.
        if np.prod(mon_size) * self.num_freq**2 == dJ.size and self.num_freq > 1:
            dJ = np.sum(dJ, axis=1)
        """The adjoint solver requires the objective function
        to be scalar valued with regard to objective arguments
        and position, but the function may be vector valued
        with regard to frequency. In this case, the Jacobian
        will be of the form [F,F,...] where F is the number of
        frequencies. Because of linearity, we can sum across the
        second frequency dimension to calculate a frequency
        scale factor for each point (rather than a scale vector).
        """

        self.all_fouriersrcdata = self._monitor.swigobj.fourier_sourcedata(
            self.volume.swigobj, self.component, self.sim.fields, dJ
        )

        for fourier_data in self.all_fouriersrcdata:
            amp_arr = np.array(fourier_data.amp_arr).reshape(-1, self.num_freq)
            scale = amp_arr * self._adj_src_scale(include_resolution=False)

            if self.num_freq == 1:
                sources += [
                    mp.IndexedSource(
                        time_src, fourier_data, scale[:, 0], not self.yee_grid
                    )
                ]
            else:
                src = FilteredSource(
                    time_src.frequency, self._frequencies, scale, self.sim.fields.dt
                )
                (num_basis, num_pts) = src.nodes.shape
                for basis_i in range(num_basis):
                    sources += [
                        mp.IndexedSource(
                            src.time_src_bf[basis_i],
                            fourier_data,
                            src.nodes[basis_i],
                            not self.yee_grid,
                        )
                    ]
        return sources

    def __call__(self):
        self._eval = np.array(
            [
                self.sim.get_dft_array(self._monitor, self.component, i)
                for i in range(self.num_freq)
            ]
        )
        return self._eval


class Near2FarFields(ObjectiveQuantity):
    def __init__(
        self,
        sim: mp.Simulation,
        Near2FarRegions: List[mp.Near2FarRegion],
        far_pts: List[mp.Vector3],
        decimation_factor: Optional[int] = 0,
        norm_near_fields: Optional[NearToFarData] = None,
    ):
        """ """
        super().__init__(sim)
        self.Near2FarRegions = Near2FarRegions
        self.far_pts = far_pts  # list of far pts
        self._nfar_pts = len(far_pts)
        self.decimation_factor = decimation_factor
        self.norm_near_fields = norm_near_fields

    def register_monitors(self, frequencies):
        self._frequencies = np.asarray(frequencies)
        self._monitor = self.sim.add_near2far(
            self._frequencies,
            *self.Near2FarRegions,
            decimation_factor=self.decimation_factor,
        )
        if self.norm_near_fields is not None:
            self.sim.load_minus_near2far_data(
                self._monitor,
                self.norm_near_fields,
            )
        return self._monitor

    def place_adjoint_source(self, dJ):
        time_src = self._create_time_profile()
        sources = []
        if dJ.ndim == 4:
            dJ = np.sum(dJ, axis=0)

        farpt_list = np.array([list(pi) for pi in self.far_pts]).flatten()
        far_pt0 = self.far_pts[0]
        far_pt_vec = py_v3_to_vec(
            self.sim.dimensions,
            far_pt0,
            self.sim.is_cylindrical,
        )

        all_nearsrcdata = self._monitor.swigobj.near_sourcedata(
            far_pt_vec, farpt_list, self._nfar_pts, dJ
        )
        for near_data in all_nearsrcdata:
            cur_comp = near_data.near_fd_comp
            amp_arr = np.array(near_data.amp_arr).reshape(-1, self.num_freq)
            scale = amp_arr * self._adj_src_scale(include_resolution=False)

            if self.num_freq == 1:
                sources += [mp.IndexedSource(time_src, near_data, scale[:, 0])]
            else:
                src = FilteredSource(
                    time_src.frequency,
                    self._frequencies,
                    scale,
                    self.sim.fields.dt,
                )
                (num_basis, num_pts) = src.nodes.shape
                for basis_i in range(num_basis):
                    sources += [
                        mp.IndexedSource(
                            src.time_src_bf[basis_i],
                            near_data,
                            src.nodes[basis_i],
                        )
                    ]
        return sources

    def __call__(self):
        self._eval = np.array(
            [self.sim.get_farfield(self._monitor, far_pt) for far_pt in self.far_pts]
        ).reshape((self._nfar_pts, self.num_freq, 6))
        return self._eval


class LDOS(ObjectiveQuantity):
    def __init__(
        self, sim: mp.Simulation, decimation_factor: Optional[int] = 0, **kwargs
    ):
        """ """
        super().__init__(sim)
        self.decimation_factor = decimation_factor
        self.srckwarg = kwargs

    def register_monitors(self, frequencies):
        self._frequencies = np.asarray(frequencies)
        self.forward_src = self.sim.sources
        return

    def place_adjoint_source(self, dJ):
        time_src = self._create_time_profile()
        if dJ.ndim == 2:
            dJ = np.sum(dJ, axis=1)
        dJ = dJ.flatten()
        sources = []
        forward_f_scale = np.array(
            [self.ldos_scale / self.ldos_Jdata[k] for k in range(self.num_freq)]
        )
        if self._frequencies.size == 1:
            amp = (dJ * self._adj_src_scale(False) * forward_f_scale)[0]
            src = time_src
        else:
            scale = dJ * self._adj_src_scale(False) * forward_f_scale
            src = FilteredSource(
                time_src.frequency,
                self._frequencies,
                scale,
                self.sim.fields.dt,
            )
            amp = 1
        for forward_src_i in self.forward_src:
            if isinstance(forward_src_i, mp.EigenModeSource):
                src_i = mp.EigenModeSource(
                    src,
                    component=forward_src_i.component,
                    eig_kpoint=forward_src_i.eig_kpoint,
                    amplitude=amp,
                    eig_band=forward_src_i.eig_band,
                    size=forward_src_i.size,
                    center=forward_src_i.center,
                    **self.srckwarg,
                )
            else:
                src_i = mp.Source(
                    src,
                    component=forward_src_i.component,
                    amplitude=amp,
                    size=forward_src_i.size,
                    center=forward_src_i.center,
                    **self.srckwarg,
                )
            if mp.is_electric(src_i.component):
                src_i.amplitude *= -1
            sources += [src_i]

        return sources

    def __call__(self):
        self._eval = self.sim.ldos_data
        self.ldos_scale = self.sim.ldos_scale
        self.ldos_Jdata = self.sim.ldos_Jdata
        return np.array(self._eval)



import matplotlib.pyplot as plt
import numpy as np

import meep as mp

# set up the geometry
eps = 13
w = 1.2
r = 0.36
d = 1.4
N = 3
sy = 6
pad = 2
dpml = 1
sx = (2 * (pad + dpml + N)) + d - 1
fcen = 0.25
df = 0.2
nfreq = 500

cell = mp.Vector3(sx, sy, 0)

blk = mp.Block(size=mp.Vector3(mp.inf, w, mp.inf), material=mp.Medium(epsilon=eps))

geometry = [blk]

geometry.extend(mp.Cylinder(r, center=mp.Vector3(d / 2 + i)) for i in range(3))
geometry.extend(mp.Cylinder(r, center=mp.Vector3(d / -2 - i)) for i in range(3))

sim = mp.Simulation(
    cell_size=cell,
    geometry=geometry,
    sources=[],
    boundary_layers=[mp.PML(dpml)],
    resolution=20,
)

# add sources
sim.sources = [mp.Source(mp.GaussianSource(fcen, fwidth=df), mp.Hz, mp.Vector3())]

# run until sources are finished (and no later)
sim._run_sources_until(0, [])

# get 1D and 2D array slices
xMin = -0.25 * sx
xMax = +0.25 * sx
yMin = -0.15 * sy
yMax = +0.15 * sy

# 1D slice of Hz data
size_1d = mp.Vector3(xMax - xMin)
center_1d = mp.Vector3((xMin + xMax) / 2)
slice1d = sim.get_array(mp.Volume(center_1d, size=size_1d), component=mp.Hz)

# 2D slice of Hz data
size_2d = mp.Vector3(xMax - xMin, yMax - yMin)
center_2d = mp.Vector3((xMin + xMax) / 2, (yMin + yMax) / 2)
slice2d = sim.get_array(mp.Volume(center_2d, size=size_2d), component=mp.Hz)

# plot 1D slice
plt.subplot(1, 2, 1)
x1d = np.linspace(xMin, xMax, len(slice1d))
plt.plot(x1d, slice1d)

# plot 2D slice
plt.subplot(1, 2, 2)
dy = (yMax - yMin) / slice2d.shape[1]
dx = (xMax - xMin) / slice2d.shape[0]
(x2d, y2d) = np.mgrid[slice(xMin, xMax, dx), slice(yMin, yMax, dy)]
plt.contourf(x2d, y2d, slice2d)
plt.colorbar()
plt.show()



import math
import unittest

import numpy as np
from utils import ApproxComparisonTestCase

import meep as mp


class TestAntennaRadiation(ApproxComparisonTestCase):
    @classmethod
    def setUp(cls):
        cls.resolution = 100  # pixels/μm

        cls.h = 1.125  # height of point source above ground plane
        cls.n = 1.2  # refractive index of surrounding medium

        cls.src_cmpt = mp.Ez
        cls.wvl = 0.65

        cls.npts = 50  # number of points in [0,pi/2) range of angles
        cls.angles = 0.5 * math.pi / cls.npts * np.arange(cls.npts)
        cls.r = 1000 * cls.wvl  # radius of far-field hemicircle

    def radial_flux(self, sim, nearfield_box, r):
        E = np.zeros((self.npts, 3), dtype=np.complex128)
        H = np.zeros((self.npts, 3), dtype=np.complex128)
        for n in range(self.npts):
            ff = sim.get_farfield(
                nearfield_box,
                mp.Vector3(r * math.sin(self.angles[n]), r * math.cos(self.angles[n])),
            )
            E[n, :] = [np.conj(ff[j]) for j in range(3)]
            H[n, :] = [ff[j + 3] for j in range(3)]

        Px = np.real(E[:, 1] * H[:, 2] - E[:, 2] * H[:, 1])  # Ey*Hz-Ez*Hy
        Py = np.real(E[:, 2] * H[:, 0] - E[:, 0] * H[:, 2])  # Ez*Hx-Ex*Hz
        return np.sqrt(np.square(Px) + np.square(Py))

    def free_space_radiation(self):
        sxy = 4
        dpml = 1
        cell_size = mp.Vector3(sxy + 2 * dpml, sxy + 2 * dpml)

        pml_layers = [mp.PML(dpml)]

        fcen = 1 / self.wvl

        sources = [
            mp.Source(
                src=mp.GaussianSource(fcen, fwidth=0.2 * fcen),
                center=mp.Vector3(),
                component=self.src_cmpt,
            )
        ]

        if self.src_cmpt == mp.Hz:
            symmetries = [mp.Mirror(mp.X, phase=-1), mp.Mirror(mp.Y, phase=-1)]
        elif self.src_cmpt == mp.Ez:
            symmetries = [mp.Mirror(mp.X, phase=+1), mp.Mirror(mp.Y, phase=+1)]
        else:
            symmetries = []

        sim = mp.Simulation(
            cell_size=cell_size,
            resolution=self.resolution,
            sources=sources,
            symmetries=symmetries,
            boundary_layers=pml_layers,
            default_material=mp.Medium(index=self.n),
        )

        nearfield_box = sim.add_near2far(
            fcen,
            0,
            1,
            mp.Near2FarRegion(
                center=mp.Vector3(0, +0.5 * sxy), size=mp.Vector3(sxy, 0)
            ),
            mp.Near2FarRegion(
                center=mp.Vector3(0, -0.5 * sxy), size=mp.Vector3(sxy, 0), weight=-1
            ),
            mp.Near2FarRegion(
                center=mp.Vector3(+0.5 * sxy, 0), size=mp.Vector3(0, sxy)
            ),
            mp.Near2FarRegion(
                center=mp.Vector3(-0.5 * sxy, 0), size=mp.Vector3(0, sxy), weight=-1
            ),
        )

        sim.run(until_after_sources=mp.stop_when_dft_decayed())

        return self.radial_flux(sim, nearfield_box, self.r)

    def pec_ground_plane_radiation(self):
        L = 8.0  # length of non-PML region
        dpml = 1.0  # thickness of PML
        sxy = dpml + L + dpml
        cell_size = mp.Vector3(sxy, sxy, 0)

        boundary_layers = [mp.PML(dpml)]

        fcen = 1 / self.wvl

        # The near-to-far field transformation feature only supports
        # homogeneous media which means it cannot explicitly take into
        # account the ground plane. As a workaround, we use two antennas
        # of _opposite_ sign surrounded by a single near2far box which
        # encloses both antennas. We then use an odd mirror symmetry to
        # cut the computational cell in half which is effectively
        # equivalent to a PEC boundary condition on one side.
        # Note: This setup means that the radiation pattern can only
        # be measured in the top half above the dipole.
        sources = [
            mp.Source(
                src=mp.GaussianSource(fcen, fwidth=0.2 * fcen),
                component=self.src_cmpt,
                center=mp.Vector3(0, +self.h),
            ),
            mp.Source(
                src=mp.GaussianSource(fcen, fwidth=0.2 * fcen),
                component=self.src_cmpt,
                center=mp.Vector3(0, -self.h),
                amplitude=-1 if self.src_cmpt == mp.Ez else +1,
            ),
        ]

        symmetries = [
            mp.Mirror(direction=mp.X, phase=+1 if self.src_cmpt == mp.Ez else -1),
            mp.Mirror(direction=mp.Y, phase=-1 if self.src_cmpt == mp.Ez else +1),
        ]

        sim = mp.Simulation(
            resolution=self.resolution,
            cell_size=cell_size,
            boundary_layers=boundary_layers,
            sources=sources,
            symmetries=symmetries,
            default_material=mp.Medium(index=self.n),
        )

        nearfield_box = sim.add_near2far(
            fcen,
            0,
            1,
            mp.Near2FarRegion(
                center=mp.Vector3(0, 2 * self.h), size=mp.Vector3(2 * self.h, 0)
            ),
            mp.Near2FarRegion(
                center=mp.Vector3(0, -2 * self.h),
                size=mp.Vector3(2 * self.h, 0),
                weight=-1,
            ),
            mp.Near2FarRegion(
                center=mp.Vector3(self.h, 0), size=mp.Vector3(0, 4 * self.h)
            ),
            mp.Near2FarRegion(
                center=mp.Vector3(-self.h, 0), size=mp.Vector3(0, 4 * self.h), weight=-1
            ),
        )

        sim.run(until_after_sources=mp.stop_when_dft_decayed())

        return self.radial_flux(sim, nearfield_box, self.r)

    def test_pec_ground_plane(self):
        """Unit test for near-to-far field transformation and symmetries.

        Verifies that the radiation pattern for a point dipole source a
        given height above a perfect-electric conductor (PEC) ground plane
        agrees with the theoretical result.

        The radiation pattern of a two-element antenna array is equivalent
        to the radiation pattern of a single antenna multiplied by its array
        factor as derived in Section 6.2 "Two-Element Array" of Antenna Theory:
        Analysis and Design, Fourth Edition (2016) by C.A. Balanis.
        """
        Pr_fsp = self.free_space_radiation()
        Pr_pec = self.pec_ground_plane_radiation()

        k = 2 * np.pi / (self.wvl / self.n)  # wavevector in medium
        Pr_theory = np.zeros(
            self.npts,
        )
        for i, ang in enumerate(self.angles):
            Pr_theory[i] = Pr_fsp[i] * 2 * np.sin(k * self.h * np.cos(ang))

        Pr_pec_norm = Pr_pec / np.max(Pr_pec)
        Pr_theory_norm = (Pr_theory / max(Pr_theory)) ** 2

        tol = 0.02
        self.assertClose(Pr_pec_norm, Pr_theory_norm, epsilon=tol)

    def test_poynting_theorem(self):
        """Unit test for near-to-far field transformation in 2d.

        Verifies that the Poynting flux of an Ez-polarized point
        dipole source in vacuum is independent of the shape of the
        enclosing measurement box due to Poynting's theorem by
        considering three arrangements:
             (1) bounding box of thenear fields,
             (2) bounding circle of the far fields, and
             (3) bounding box of the far fields.
        """
        resolution = 50
        sxy = 4
        dpml = 1
        cell = mp.Vector3(sxy + 2 * dpml, sxy + 2 * dpml)

        pml_layers = mp.PML(dpml)

        fcen = 1.0
        df = 0.4

        sources = [
            mp.Source(
                src=mp.GaussianSource(fcen, fwidth=df),
                center=mp.Vector3(),
                component=mp.Ez,
            )
        ]

        symmetries = [mp.Mirror(mp.X), mp.Mirror(mp.Y)]

        sim = mp.Simulation(
            cell_size=cell,
            resolution=resolution,
            sources=sources,
            symmetries=symmetries,
            boundary_layers=[pml_layers],
        )

        nearfield_box = sim.add_near2far(
            fcen,
            0,
            1,
            mp.Near2FarRegion(mp.Vector3(y=0.5 * sxy), size=mp.Vector3(sxy)),
            mp.Near2FarRegion(
                mp.Vector3(y=-0.5 * sxy), size=mp.Vector3(sxy), weight=-1
            ),
            mp.Near2FarRegion(mp.Vector3(0.5 * sxy), size=mp.Vector3(y=sxy)),
            mp.Near2FarRegion(
                mp.Vector3(-0.5 * sxy), size=mp.Vector3(y=sxy), weight=-1
            ),
        )

        flux_box = sim.add_flux(
            fcen,
            0,
            1,
            mp.FluxRegion(mp.Vector3(y=0.5 * sxy), size=mp.Vector3(sxy)),
            mp.FluxRegion(mp.Vector3(y=-0.5 * sxy), size=mp.Vector3(sxy), weight=-1),
            mp.FluxRegion(mp.Vector3(0.5 * sxy), size=mp.Vector3(y=sxy)),
            mp.FluxRegion(mp.Vector3(-0.5 * sxy), size=mp.Vector3(y=sxy), weight=-1),
        )

        sim.run(
            until_after_sources=mp.stop_when_fields_decayed(
                50, mp.Ez, mp.Vector3(), 1e-8
            )
        )

        near_flux = mp.get_fluxes(flux_box)[0]

        r = 1000 / fcen  # radius of far field circle
        Pr = self.radial_flux(sim, nearfield_box, r)
        far_flux_circle = 4 * np.sum(Pr) * 0.5 * np.pi * r / len(Pr)

        rr = 20 / fcen  # length of far-field square box
        res_far = 20  # resolution of far-field square box
        far_flux_square = (
            nearfield_box.flux(
                mp.Y,
                mp.Volume(center=mp.Vector3(y=0.5 * rr), size=mp.Vector3(rr)),
                res_far,
            )[0]
            - nearfield_box.flux(
                mp.Y,
                mp.Volume(center=mp.Vector3(y=-0.5 * rr), size=mp.Vector3(rr)),
                res_far,
            )[0]
            + nearfield_box.flux(
                mp.X,
                mp.Volume(center=mp.Vector3(0.5 * rr), size=mp.Vector3(y=rr)),
                res_far,
            )[0]
            - nearfield_box.flux(
                mp.X,
                mp.Volume(center=mp.Vector3(-0.5 * rr), size=mp.Vector3(y=rr)),
                res_far,
            )[0]
        )

        print(
            "flux:, {:.6f}, {:.6f}, {:.6f}".format(
                near_flux, far_flux_circle, far_flux_square
            )
        )

        self.assertAlmostEqual(near_flux, far_flux_circle, places=2)
        self.assertAlmostEqual(far_flux_circle, far_flux_square, places=2)
        self.assertAlmostEqual(far_flux_square, near_flux, places=2)


if __name__ == "__main__":
    unittest.main()



# Calculating 2d ring-resonator modes, from the Meep tutorial.
import meep as mp


def main():
    n = 3.4  # index of waveguide
    w = 1  # width of waveguide
    r = 1  # inner radius of ring
    pad = 4  # padding between waveguide and edge of PML
    dpml = 2  # thickness of PML
    sxy = 2 * (r + w + pad + dpml)  # cell size

    # Create a ring waveguide by two overlapping cylinders - later objects
    # take precedence over earlier objects, so we put the outer cylinder first.
    # and the inner (air) cylinder second.

    c1 = mp.Cylinder(radius=r + w, material=mp.Medium(index=n))
    c2 = mp.Cylinder(radius=r)

    # If we don't want to excite a specific mode symmetry, we can just
    # put a single point source at some arbitrary place, pointing in some
    # arbitrary direction.  We will only look for Ez-polarized modes.

    fcen = 0.15  # pulse center frequency
    df = 0.1  # pulse width (in frequency)

    src = mp.Source(mp.GaussianSource(fcen, fwidth=df), mp.Ez, mp.Vector3(r + 0.1))

    sim = mp.Simulation(
        cell_size=mp.Vector3(sxy, sxy),
        geometry=[c1, c2],
        sources=[src],
        resolution=10,
        symmetries=[mp.Mirror(mp.Y)],
        boundary_layers=[mp.PML(dpml)],
    )

    sim.run(
        mp.at_beginning(mp.output_epsilon),
        mp.after_sources(mp.Harminv(mp.Ez, mp.Vector3(r + 0.1), fcen, df)),
        until_after_sources=300,
    )

    # Output fields for one period at the end.  (If we output
    # at a single time, we might accidentally catch the Ez field when it is
    # almost zero and get a distorted view.)
    sim.run(mp.at_every(1 / fcen / 20, mp.output_efield_z), until=1 / fcen)


if __name__ == "__main__":
    main()



import meep as mp

try:
    import meep.adjoint as mpa
except:
    import adjoint as mpa

import unittest
from enum import Enum

import numpy as np
from autograd import numpy as npa
from autograd import tensor_jacobian_product
from utils import ApproxComparisonTestCase
import parameterized

rng = np.random.RandomState(2)
resolution = 20
dimensions = mp.CYLINDRICAL
Si = mp.Medium(index=3.4)
SiO2 = mp.Medium(index=1.44)

sr = 6
sz = 6
cell_size = mp.Vector3(sr, 0, sz)
dpml = 1.0
boundary_layers = [mp.PML(thickness=dpml)]

design_region_resolution = int(2 * resolution)
design_r = 5
design_z = 2
Nr, Nz = int(design_r * design_region_resolution), int(
    design_z * design_region_resolution
)

fcen = 1 / 1.55
width = 0.2
fwidth = width * fcen
source_center = [design_r / 2, 0, -(sz / 2 - dpml + design_z / 2) / 2]
source_size = mp.Vector3(design_r, 0, 0)
src = mp.GaussianSource(frequency=fcen, fwidth=fwidth)
source = [mp.Source(src, component=mp.Er, center=source_center, size=source_size)]

## random design region
p = 0.5 * rng.rand(Nr * Nz)
## random epsilon perturbation for design region
deps = 1e-5
dp = deps * rng.rand(Nr * Nz)


def forward_simulation(design_params, m, far_x):
    matgrid = mp.MaterialGrid(
        mp.Vector3(Nr, 0, Nz), SiO2, Si, weights=design_params.reshape(Nr, 1, Nz)
    )

    geometry = [
        mp.Block(
            center=mp.Vector3(design_r / 2, 0, 0),
            size=mp.Vector3(design_r, 0, design_z),
            material=matgrid,
        )
    ]

    sim = mp.Simulation(
        resolution=resolution,
        cell_size=cell_size,
        boundary_layers=boundary_layers,
        sources=source,
        geometry=geometry,
        dimensions=dimensions,
        m=m,
    )

    frequencies = [fcen]

    mode = sim.add_near2far(
        frequencies,
        mp.Near2FarRegion(
            center=mp.Vector3(design_r / 2, 0, (sz / 2 - dpml + design_z / 2) / 2),
            size=mp.Vector3(design_r, 0, 0),
            weight=+1,
        ),
    )

    sim.run(until_after_sources=1200)
    Er = sim.get_farfield(mode, far_x[0])
    sim.reset_meep()

    return abs(Er[0]) ** 2


def adjoint_solver(design_params, m, far_x):

    design_variables = mp.MaterialGrid(mp.Vector3(Nr, 0, Nz), SiO2, Si)
    design_region = mpa.DesignRegion(
        design_variables,
        volume=mp.Volume(
            center=mp.Vector3(design_r / 2, 0, 0),
            size=mp.Vector3(design_r, 0, design_z),
        ),
    )
    geometry = [
        mp.Block(
            center=design_region.center,
            size=design_region.size,
            material=design_variables,
        )
    ]

    sim = mp.Simulation(
        cell_size=cell_size,
        boundary_layers=boundary_layers,
        geometry=geometry,
        sources=source,
        resolution=resolution,
        dimensions=dimensions,
        m=m,
    )

    NearRegions = [
        mp.Near2FarRegion(
            center=mp.Vector3(design_r / 2, 0, (sz / 2 - dpml + design_z / 2) / 2),
            size=mp.Vector3(design_r, 0, 0),
            weight=+1,
        )
    ]
    FarFields = mpa.Near2FarFields(sim, NearRegions, far_x)
    ob_list = [FarFields]

    def J(alpha):
        return npa.abs(alpha[0, 0, 0]) ** 2

    opt = mpa.OptimizationProblem(
        simulation=sim,
        objective_functions=J,
        objective_arguments=ob_list,
        design_regions=[design_region],
        fcen=fcen,
        df=0,
        nf=1,
        maximum_run_time=1200,
    )

    f, dJ_du = opt([design_params])
    sim.reset_meep()

    return f, dJ_du


class TestAdjointSolver(ApproxComparisonTestCase):
    @parameterized.parameterized.expand(
        [
            (0, [mp.Vector3(5, 0, 20)]),
            (0, [mp.Vector3(4, 0, 28)]),
            (-1, [mp.Vector3(5, 0, 20)]),
            (1.2, [mp.Vector3(5, 0, 20)]),
        ]
    )
    def test_adjoint_solver_cyl_n2f_fields(self, m, far_x):
        print("*** TESTING CYLINDRICAL Near2Far ADJOINT FEATURES ***")
        print(f"Current test: m={m}, far_x={far_x}")
        adjsol_obj, adjsol_grad = adjoint_solver(p, m, far_x)

        ## compute unperturbed S12
        S12_unperturbed = forward_simulation(p, m, far_x)

        ## compare objective results
        print(
            f"|Er|^2 -- adjoint solver: {adjsol_obj}, traditional simulation: {S12_unperturbed}"
        )

        self.assertClose(adjsol_obj, S12_unperturbed, epsilon=1e-3)

        ## compute perturbed S12
        S12_perturbed = forward_simulation(p + dp, m, far_x)

        ## compare gradients
        if adjsol_grad.ndim < 2:
            adjsol_grad = np.expand_dims(adjsol_grad, axis=1)
        adj_scale = (dp[None, :] @ adjsol_grad).flatten()
        fd_grad = S12_perturbed - S12_unperturbed
        print(f"Directional derivative -- adjoint solver: {adj_scale}, FD: {fd_grad}")
        tol = 0.2 if mp.is_single_precision() else 0.1
        self.assertClose(adj_scale, fd_grad, epsilon=tol)


if __name__ == "__main__":
    unittest.main()



from abc import ABCMeta, abstractmethod

import numpy as np
from scipy import sparse

import meep as mp

ABC = ABCMeta("ABC", (object,), {"__slots__": ()})  # compatible with Python 2 and 3


# ----------------------------------------------------------------------
# Basis is the abstract base class from which classes describing specific
# basis sets should inherit.
# ----------------------------------------------------------------------
class Basis(ABC):
    """ """

    def __init__(
        self,
        rho_vector=None,
        volume=None,
        size=None,
        center=mp.Vector3(),
    ):
        self.volume = volume or mp.Volume(center=center, size=size)
        self.rho_vector = rho_vector

    def func(self):
        def _f(p):
            return self(p)

        return _f

    @abstractmethod
    def get_basis_vjp(self):
        raise NotImplementedError("derived class must implement get_basis_vjp method")

    @abstractmethod
    def __call__(self, p=[0.0, 0.0]):
        raise NotImplementedError("derived class must implement __call__() method")

    def set_rho_vector(self, rho_vector):
        self.rho_vector = rho_vector


# -------------------------------- #
# Bilinear Interpolation Basis class
# -------------------------------- #


class BilinearInterpolationBasis(Basis):
    """
    Simple bilinear interpolation basis set.
    """

    def __init__(self, resolution, symmetry=None, **kwargs):
        self.dim = 2

        super().__init__(**kwargs)

        # Generate interpolation grid
        self.symmetry = [] if symmetry is None or len(symmetry) == 0 else symmetry
        if mp.X in set(self.symmetry):
            self.Nx = int(resolution * self.volume.size.x / 2) + 1
            self.rho_x = np.linspace(
                self.volume.center.x,
                self.volume.center.x + self.volume.size.x / 2,
                self.Nx,
            )
            self.mirror_X = True
        else:
            self.Nx = int(resolution * self.volume.size.x) + 1
            self.rho_x = np.linspace(
                self.volume.center.x - self.volume.size.x / 2,
                self.volume.center.x + self.volume.size.x / 2,
                self.Nx,
            )
            self.mirror_X = False

        if mp.Y in set(self.symmetry):
            self.Ny = int(resolution * self.volume.size.y / 2) + 1
            self.rho_y = np.linspace(
                self.volume.center.y,
                self.volume.center.y + self.volume.size.y / 2,
                self.Ny,
            )
            self.mirror_Y = True
        else:
            self.Ny = int(resolution * self.volume.size.y) + 1
            self.rho_y = np.linspace(
                self.volume.center.y - self.volume.size.y / 2,
                self.volume.center.y + self.volume.size.y / 2,
                self.Ny,
            )
            self.mirror_Y = False

        self.num_design_params = self.Nx * self.Ny

        if self.rho_vector is None:
            self.rho_vector = np.ones((self.num_design_params,))

    def __call__(self, p):
        x = (
            2 * self.volume.center.x - p.x
            if self.mirror_X and p.x < self.volume.center.x
            else p.x
        )
        y = (
            2 * self.volume.center.y - p.y
            if self.mirror_Y and p.y < self.volume.center.y
            else p.y
        )

        weights, interp_idx = self.get_bilinear_row(
            x, y, self.rho_x, self.rho_y
        )  # ignore z coordinate
        return np.dot(self.rho_vector[interp_idx], weights)

    def get_basis_vjp(self, dJ_deps, design_grid):
        """get vector jacobian product of interpolator"""

        dg_Nx, dg_Ny, Nz, Nf = dJ_deps.shape  # get important design grid dimensions
        x_grid = design_grid.x
        y_grid = design_grid.y
        z_grid = design_grid.z

        # take care of symmetries
        if self.mirror_X:
            dJ_deps = dJ_deps[int(dg_Nx / 2) :, :, :, :] * 2
            x_grid = x_grid[int(dg_Nx / 2) :]
        if self.mirror_Y:
            dJ_deps = dJ_deps[:, int(dg_Ny / 2) :, :, :] * 2
            y_grid = y_grid[int(dg_Ny / 2) :]

        dg_Nx, dg_Ny, Nz, Nf = dJ_deps.shape  # recalculate
        Nx, Ny = (
            self.rho_x.size,
            self.rho_y.size,
        )  # get important interpolator dimensions

        # same interpolation matrix for all frequencies and all coordinates in Z direction
        A = self.gen_interpolation_matrix(
            self.rho_x, self.rho_y, x_grid, y_grid, z_grid
        )
        # TODO ditch the for loops
        dJ_dp = np.zeros((Nx * Ny, Nf))
        for fi in range(Nf):
            for zi in range(Nz):
                dJ_dp[:, fi] += np.matmul(
                    dJ_deps[:, :, zi, fi].reshape(dg_Nx * dg_Ny, order="C"), A
                )
        return dJ_dp

    def get_bilinear_coefficients(self, x, x1, x2, y, y1, y2):
        """
        Calculates the bilinear interpolation coefficients for a single point at (x,y).
        Assumes that the user already knows the four closest points and provides the corresponding
        (x1,x2) and(y1,y2) coordinates.
        """
        b11 = ((x - x2) * (y - y2)) / ((x1 - x2) * (y1 - y2))
        b12 = -((x - x2) * (y - y1)) / ((x1 - x2) * (y1 - y2))
        b21 = -((x - x1) * (y - y2)) / ((x1 - x2) * (y1 - y2))
        b22 = ((x - x1) * (y - y1)) / ((x1 - x2) * (y1 - y2))
        return [b11, b12, b21, b22]

    def get_bilinear_row(self, rx, ry, rho_x, rho_y):
        """
        Calculates a vector of bilinear interpolation weights that can be used
        in an inner product with the neighboring function values, or placed
        inside of an interpolation matrix.
        """

        Nx = rho_x.size
        Ny = rho_y.size

        # binary search in x direction to get x1 and x2
        xi2 = np.searchsorted(rho_x, rx, side="left")
        if xi2 <= 0:  # extrapolation (be careful!)
            xi1 = 0
            xi2 = 1
        elif xi2 >= Nx - 1:  # extrapolation (be careful!)
            xi1 = Nx - 2
            xi2 = Nx - 1
        else:
            xi1 = xi2 - 1

        x1 = rho_x[xi1]
        x2 = rho_x[xi2]

        # binary search in y direction to get y1 and y2
        yi2 = np.searchsorted(rho_y, ry, side="left")
        if yi2 <= 0:  # extrapolation (be careful!)
            yi1 = 0
            yi2 = 1
        elif yi2 >= Ny - 1:  # extrapolation (be careful!)
            yi1 = Ny - 2
            yi2 = Ny - 1
        else:
            yi1 = yi2 - 1

        y1 = rho_y[yi1]
        y2 = rho_y[yi2]

        # get weights
        weights = self.get_bilinear_coefficients(rx, x1, x2, ry, y1, y2)

        # get location of nearest neigbor interpolation points
        interp_idx = np.array(
            [xi1 * Ny + yi1, xi1 * Ny + yi2, (xi2) * Ny + yi1, (xi2) * Ny + yi2],
            dtype=np.int64,
        )

        return weights, interp_idx

    def gen_interpolation_matrix(
        self,
        rho_x,
        rho_y,
        rho_x_interp,
        rho_y_interp,
        rho_z_interp,
    ):
        """
        Generates a bilinear interpolation matrix.

        Arguments:
        rho_x ................ [N,] numpy array - original x array mapping to povided data
        rho_y ................ [N,] numpy array - original y array mapping to povided data
        rho_x_interp ......... [N,] numpy array - new x array mapping to desired interpolated data
        rho_y_interp ......... [N,] numpy array - new y array mapping to desired interpolated data

        Returns:
        A .................... [N,M] sparse matrix - interpolation matrix
        """

        Nx = rho_x.size
        Ny = rho_y.size
        Nx_interp = np.array(rho_x_interp).size
        Ny_interp = np.array(rho_y_interp).size
        Nz_interp = np.array(rho_y_interp).size

        input_dimension = Nx * Ny
        output_dimension = Nx_interp * Ny_interp

        interp_weights = np.zeros(4 * output_dimension)
        row_ind = np.zeros(4 * output_dimension, dtype=np.int64)
        col_ind = np.zeros(4 * output_dimension, dtype=np.int64)

        ri = 0
        for rx in rho_x_interp:
            for ry in rho_y_interp:
                # get weights
                weights, interp_idx = self.get_bilinear_row(rx, ry, rho_x, rho_y)

                # populate sparse matrix vectors
                interp_weights[4 * ri : 4 * (ri + 1)] = weights
                row_ind[4 * ri : 4 * (ri + 1)] = np.array(
                    [ri, ri, ri, ri], dtype=np.int64
                )
                col_ind[4 * ri : 4 * (ri + 1)] = interp_idx

                ri += 1

        return sparse.coo_matrix(
            (interp_weights, (row_ind, col_ind)),
            shape=(output_dimension, input_dimension),
        )



import copy
import unittest

import meep.binary_partition_utils as bpu
import numpy as np
import parameterized

import meep as mp

PARTITION_NO_DUPLICATE_PROC_ID = mp.BinaryPartition(
    data=[(mp.X, -2.0), 0, [(mp.Y, 1.5), [(mp.X, 4.0), 1, [(mp.Y, 0.5), 4, 3]], 2]]
)
PARTITION_DUPLICATE_PROC_ID = mp.BinaryPartition(
    data=[(mp.X, -2.0), 0, [(mp.Y, 1.5), [(mp.X, 4.0), 1, [(mp.Y, 0.5), 1, 3]], 0]]
)

# Mocked chunk IDs since we are in a single-core environment
CHUNK_OWNERS_NO_DUPLICATE_PROC_ID = np.array([0, 1, 4, 3, 2])
CHUNK_OWNERS_DUPLICATE_PROC_ID = np.array([0, 1, 1, 3, 0])


class BinaryPartitionUtilsTest(unittest.TestCase):
    @parameterized.parameterized.expand(
        [
            (PARTITION_NO_DUPLICATE_PROC_ID, False),
            (PARTITION_NO_DUPLICATE_PROC_ID.right, False),
            (PARTITION_NO_DUPLICATE_PROC_ID.left, True),
            (PARTITION_DUPLICATE_PROC_ID, False),
            (PARTITION_DUPLICATE_PROC_ID.right, False),
            (PARTITION_DUPLICATE_PROC_ID.left, True),
        ]
    )
    def test_is_leaf_node(self, partition, expected_leaf_status):
        self.assertEqual(bpu.is_leaf_node(partition), expected_leaf_status)

    @parameterized.parameterized.expand(
        [
            (PARTITION_NO_DUPLICATE_PROC_ID, CHUNK_OWNERS_NO_DUPLICATE_PROC_ID),
            (PARTITION_DUPLICATE_PROC_ID, CHUNK_OWNERS_DUPLICATE_PROC_ID),
        ]
    )
    def test_enumerate_leaf_nodes(self, partition, chunk_owners):
        leaf_nodes = list(bpu.enumerate_leaf_nodes(partition))
        self.assertEqual(len(leaf_nodes), partition.numchunks())
        proc_ids = [node.proc_id for node in bpu.enumerate_leaf_nodes(partition)]
        self.assertEqual(proc_ids, list(chunk_owners))  # depth first ordering

    def test_partition_has_duplicate_proc_ids(self):
        self.assertFalse(
            bpu.partition_has_duplicate_proc_ids(PARTITION_NO_DUPLICATE_PROC_ID)
        )
        self.assertTrue(
            bpu.partition_has_duplicate_proc_ids(PARTITION_DUPLICATE_PROC_ID)
        )

    @parameterized.parameterized.expand(
        [
            (
                PARTITION_NO_DUPLICATE_PROC_ID,
                [0, 1, 2, 3, 4],
            ),
            (
                PARTITION_DUPLICATE_PROC_ID,
                [0, 1, 2, 3, 4],
            ),
        ]
    )
    def test_get_total_weight(self, partition, weights_by_proc_id):
        if not bpu.partition_has_duplicate_proc_ids(partition):
            self.assertEqual(
                bpu.get_total_weight(partition, weights_by_proc_id),
                sum(weights_by_proc_id),
            )
            self.assertEqual(
                bpu.get_total_weight(partition.right.left, weights_by_proc_id),
                weights_by_proc_id[1] + weights_by_proc_id[4] + weights_by_proc_id[3],
            )
        else:
            with self.assertRaises(ValueError):
                bpu.get_total_weight(partition, weights_by_proc_id)

    @parameterized.parameterized.expand(
        [
            (
                PARTITION_NO_DUPLICATE_PROC_ID,
                [0, 1, 2, 3, 4],
            ),
            (
                PARTITION_DUPLICATE_PROC_ID,
                [0, 1, 2, 3, 4],
            ),
        ]
    )
    def test_get_left_right_total_weights(self, partition, weights_by_proc_id):
        proc_ids = [node.proc_id for node in bpu.enumerate_leaf_nodes(partition)]
        no_duplicates = len(set(proc_ids)) == len(proc_ids)
        if no_duplicates:
            self.assertEqual(
                bpu.get_left_right_total_weights(partition, weights_by_proc_id),
                (
                    bpu.get_total_weight(partition.left, weights_by_proc_id),
                    bpu.get_total_weight(partition.right, weights_by_proc_id),
                ),
            )
        else:
            with self.assertRaises(ValueError):
                bpu.get_left_right_total_weights(partition, weights_by_proc_id)

    @parameterized.parameterized.expand(
        [
            (
                PARTITION_NO_DUPLICATE_PROC_ID,
                2,
                [1500, 2400, 300, 100, 700],
            ),
            (
                PARTITION_NO_DUPLICATE_PROC_ID,
                3,
                [15000, 24000, 3000, 1000, 7000],
            ),
        ]
    )
    def test_pixel_volume(self, partition, dims, expected_pixel_volumes):
        cell_size = (
            mp.Vector3(10.0, 5.0, 1.0) if dims == 3 else mp.Vector3(10.0, 5.0, 0.0)
        )
        sim = mp.Simulation(cell_size=cell_size, resolution=10, chunk_layout=partition)
        sim.init_sim()
        chunk_volumes = sim.structure.get_chunk_volumes()
        self.assertEqual(
            [bpu.pixel_volume(vol) for vol in chunk_volumes], expected_pixel_volumes
        )

    @parameterized.parameterized.expand(
        [
            (PARTITION_NO_DUPLICATE_PROC_ID, CHUNK_OWNERS_NO_DUPLICATE_PROC_ID, 3500),
            (PARTITION_DUPLICATE_PROC_ID, CHUNK_OWNERS_DUPLICATE_PROC_ID, 3500),
        ]
    )
    def test_get_total_volume_2d(self, partition, chunk_owners, expected_total_volume):
        sim = mp.Simulation(
            cell_size=mp.Vector3(10.0, 5.0, 0.0), resolution=10, chunk_layout=partition
        )
        sim.init_sim()

        chunk_volumes = sim.structure.get_chunk_volumes()
        total_volume = sim.cell_size[0] * sim.cell_size[1] * sim.resolution**2
        self.assertEqual(
            bpu.get_total_volume(partition, chunk_volumes, chunk_owners), total_volume
        )

        if not bpu.partition_has_duplicate_proc_ids(partition):
            self.assertEqual(
                bpu.get_total_volume(partition.right, chunk_volumes, chunk_owners),
                expected_total_volume,
            )

    @parameterized.parameterized.expand(
        [
            (PARTITION_NO_DUPLICATE_PROC_ID, CHUNK_OWNERS_NO_DUPLICATE_PROC_ID, 35000),
            (PARTITION_DUPLICATE_PROC_ID, CHUNK_OWNERS_DUPLICATE_PROC_ID, 35000),
        ]
    )
    def test_get_total_volume_3d(self, partition, chunk_owners, expected_total_volume):
        sim = mp.Simulation(
            cell_size=mp.Vector3(10.0, 5.0, 1.0), resolution=10, chunk_layout=partition
        )
        sim.init_sim()

        chunk_volumes = sim.structure.get_chunk_volumes()
        total_volume = (
            sim.cell_size[0] * sim.cell_size[1] * sim.cell_size[2] * sim.resolution**3
        )
        self.assertEqual(
            bpu.get_total_volume(partition, chunk_volumes, chunk_owners), total_volume
        )

        if not bpu.partition_has_duplicate_proc_ids(partition):
            self.assertEqual(
                bpu.get_total_volume(partition.right, chunk_volumes, chunk_owners),
                expected_total_volume,
            )

    @parameterized.parameterized.expand(
        [
            (
                PARTITION_NO_DUPLICATE_PROC_ID,
                CHUNK_OWNERS_NO_DUPLICATE_PROC_ID,
                2,
            ),
            (
                PARTITION_DUPLICATE_PROC_ID,
                CHUNK_OWNERS_DUPLICATE_PROC_ID,
                2,
            ),
            (
                PARTITION_NO_DUPLICATE_PROC_ID,
                CHUNK_OWNERS_NO_DUPLICATE_PROC_ID,
                3,
            ),
            (
                PARTITION_DUPLICATE_PROC_ID,
                CHUNK_OWNERS_DUPLICATE_PROC_ID,
                3,
            ),
        ]
    )
    def test_get_left_right_total_volumes(self, partition, chunk_owners, dims):
        cell_size = (
            mp.Vector3(10.0, 5.0, 1.0) if dims == 3 else mp.Vector3(10.0, 5.0, 0.0)
        )
        sim = mp.Simulation(cell_size=cell_size, resolution=10, chunk_layout=partition)
        sim.init_sim()

        chunk_volumes = sim.structure.get_chunk_volumes()
        self.assertEqual(
            bpu.get_left_right_total_volumes(partition, chunk_volumes, chunk_owners),
            (
                bpu.get_total_volume(partition.left, chunk_volumes, chunk_owners),
                bpu.get_total_volume(partition.right, chunk_volumes, chunk_owners),
            ),
        )

    @parameterized.parameterized.expand(
        [
            (
                PARTITION_NO_DUPLICATE_PROC_ID,
                CHUNK_OWNERS_NO_DUPLICATE_PROC_ID,
                2,
            ),
            (
                PARTITION_DUPLICATE_PROC_ID,
                CHUNK_OWNERS_DUPLICATE_PROC_ID,
                2,
            ),
            (
                PARTITION_NO_DUPLICATE_PROC_ID,
                CHUNK_OWNERS_NO_DUPLICATE_PROC_ID,
                3,
            ),
            (
                PARTITION_DUPLICATE_PROC_ID,
                CHUNK_OWNERS_DUPLICATE_PROC_ID,
                3,
            ),
        ]
    )
    def test_get_grid_volumes_in_tree(self, partition, chunk_owners, dims):
        cell_size = (
            mp.Vector3(10.0, 5.0, 1.0) if dims == 3 else mp.Vector3(10.0, 5.0, 0.0)
        )
        sim = mp.Simulation(cell_size=cell_size, resolution=10, chunk_layout=partition)
        sim.init_sim()

        chunk_volumes = sim.structure.get_chunk_volumes()
        grid_volumes_in_tree = bpu.get_grid_volumes_in_tree(
            partition, chunk_volumes, chunk_owners
        )
        self.assertEqual(set(grid_volumes_in_tree), set(chunk_volumes))

        no_duplicates = len(set(chunk_owners)) == len(chunk_owners)
        grid_volumes_in_right_expected = (
            chunk_volumes[1:] if no_duplicates else chunk_volumes
        )
        grid_volumes_in_right = bpu.get_grid_volumes_in_tree(
            partition.right, chunk_volumes, chunk_owners
        )
        self.assertEqual(
            set(grid_volumes_in_right), set(grid_volumes_in_right_expected)
        )

    @parameterized.parameterized.expand(
        [
            (
                PARTITION_NO_DUPLICATE_PROC_ID,
                CHUNK_OWNERS_NO_DUPLICATE_PROC_ID,
                2,
                {0: 1500, 1: 2400, 4: 300, 3: 100, 2: 700},
            ),
            (
                PARTITION_DUPLICATE_PROC_ID,
                CHUNK_OWNERS_DUPLICATE_PROC_ID,
                2,
                {
                    0: 2200,
                    1: 2700,
                    3: 100,
                },
            ),
            (
                PARTITION_NO_DUPLICATE_PROC_ID,
                CHUNK_OWNERS_NO_DUPLICATE_PROC_ID,
                3,
                {0: 15000, 1: 24000, 4: 3000, 3: 1000, 2: 7000},
            ),
            (
                PARTITION_DUPLICATE_PROC_ID,
                CHUNK_OWNERS_DUPLICATE_PROC_ID,
                3,
                {
                    0: 22000,
                    1: 27000,
                    3: 1000,
                },
            ),
        ]
    )
    def test_get_total_volume_per_process(
        self, partition, chunk_owners, dims, expected_volumes_per_process
    ):
        cell_size = (
            mp.Vector3(10.0, 5.0, 1.0) if dims == 3 else mp.Vector3(10.0, 5.0, 0.0)
        )
        sim = mp.Simulation(cell_size=cell_size, resolution=10, chunk_layout=partition)
        sim.init_sim()
        chunk_volumes = sim.structure.get_chunk_volumes()
        volumes_per_process = bpu.get_total_volume_per_process(
            partition, chunk_volumes, chunk_owners
        )
        self.assertEqual(volumes_per_process, expected_volumes_per_process)

    @parameterized.parameterized.expand(
        [
            (
                PARTITION_NO_DUPLICATE_PROC_ID,
                CHUNK_OWNERS_NO_DUPLICATE_PROC_ID,
                2,
                (-5.0, 5.0, -2.5, 2.5, 0.0, 0.0),
                (-2.0, 5.0, -2.5, 2.5, 0.0, 0.0),
            ),
            (
                PARTITION_DUPLICATE_PROC_ID,
                CHUNK_OWNERS_DUPLICATE_PROC_ID,
                2,
                (-5.0, 5.0, -2.5, 2.5, 0.0, 0.0),
                (-5.0, 5.0, -2.5, 2.5, 0.0, 0.0),
            ),
            (
                PARTITION_NO_DUPLICATE_PROC_ID,
                CHUNK_OWNERS_NO_DUPLICATE_PROC_ID,
                3,
                (-5.0, 5.0, -2.5, 2.5, -0.5, 0.5),
                (-2.0, 5.0, -2.5, 2.5, -0.5, 0.5),
            ),
            (
                PARTITION_DUPLICATE_PROC_ID,
                CHUNK_OWNERS_DUPLICATE_PROC_ID,
                3,
                (-5.0, 5.0, -2.5, 2.5, -0.5, 0.5),
                (-5.0, 5.0, -2.5, 2.5, -0.5, 0.5),
            ),
        ]
    )
    def test_get_box_ranges(
        self,
        partition,
        chunk_owners,
        dims,
        expected_box_ranges,
        expected_right_box_ranges,
    ):
        cell_size = (
            mp.Vector3(10.0, 5.0, 1.0) if dims == 3 else mp.Vector3(10.0, 5.0, 0.0)
        )
        sim = mp.Simulation(cell_size=cell_size, resolution=10, chunk_layout=partition)
        sim.init_sim()
        chunk_volumes = sim.structure.get_chunk_volumes()
        self.assertEqual(
            bpu.get_box_ranges(partition, chunk_volumes, chunk_owners),
            expected_box_ranges,
        )
        self.assertEqual(
            bpu.get_box_ranges(partition.right, chunk_volumes, chunk_owners),
            expected_right_box_ranges,
        )

    @parameterized.parameterized.expand(
        [
            (
                copy.deepcopy(PARTITION_NO_DUPLICATE_PROC_ID),
                copy.deepcopy(PARTITION_NO_DUPLICATE_PROC_ID),
                True,
            ),
            (
                copy.deepcopy(PARTITION_DUPLICATE_PROC_ID),
                copy.deepcopy(PARTITION_DUPLICATE_PROC_ID),
                True,
            ),
            (
                copy.deepcopy(PARTITION_NO_DUPLICATE_PROC_ID),
                copy.deepcopy(PARTITION_DUPLICATE_PROC_ID),
                False,
            ),
            (
                mp.BinaryPartition(data=[(mp.X, -2.5), 0, [(mp.X, 2.5), 1, 2]]),
                mp.BinaryPartition(data=[(mp.X, -2.5), 0, [(mp.X, 2.5), 1, 2]]),
                True,
            ),
            (
                mp.BinaryPartition(data=[(mp.X, -2.5), 0, [(mp.X, 2.4), 1, 2]]),
                mp.BinaryPartition(data=[(mp.X, -2.5), 0, [(mp.X, 2.5), 1, 2]]),
                False,
            ),
        ]
    )
    def test_partitions_are_equal(self, bp1, bp2, is_equal):
        self.assertEqual(bpu.partitions_are_equal(bp1, bp2), is_equal)


if __name__ == "__main__":
    unittest.main()



# This file was automatically generated by SWIG (http://www.swig.org).
# Version 4.0.2
#
# Do not make changes to this file unless you know what you are doing--modify
# the SWIG interface file instead.

from sys import version_info as _swig_python_version_info
if _swig_python_version_info < (2, 7, 0):
    raise RuntimeError("Python 2.7 or later required")

# Import the low-level C/C++ module
if __package__ or "." in __name__:
    from . import _mpb
else:
    import _mpb

try:
    import builtins as __builtin__
except ImportError:
    import __builtin__

def _swig_repr(self):
    try:
        strthis = "proxy of " + self.this.__repr__()
    except __builtin__.Exception:
        strthis = ""
    return "<%s.%s; %s >" % (self.__class__.__module__, self.__class__.__name__, strthis,)


def _swig_setattr_nondynamic_instance_variable(set):
    def set_instance_attr(self, name, value):
        if name == "thisown":
            self.this.own(value)
        elif name == "this":
            set(self, name, value)
        elif hasattr(self, name) and isinstance(getattr(type(self), name), property):
            set(self, name, value)
        else:
            raise AttributeError("You cannot add instance attributes to %s" % self)
    return set_instance_attr


def _swig_setattr_nondynamic_class_variable(set):
    def set_class_attr(cls, name, value):
        if hasattr(cls, name) and not isinstance(getattr(cls, name), property):
            set(cls, name, value)
        else:
            raise AttributeError("You cannot add class attributes to %s" % cls)
    return set_class_attr


def _swig_add_metaclass(metaclass):
    """Class decorator for adding a metaclass to a SWIG wrapped class - a slimmed down version of six.add_metaclass"""
    def wrapper(cls):
        return metaclass(cls.__name__, cls.__bases__, cls.__dict__.copy())
    return wrapper


class _SwigNonDynamicMeta(type):
    """Meta class to enforce nondynamic attributes (no new attributes) for a class"""
    __setattr__ = _swig_setattr_nondynamic_class_variable(type.__setattr__)


import meep
TWOPI = _mpb.TWOPI

def map_data(d_in_re, d_in_im, n_in, d_out_re, d_out_im, n_out, coord_map, kvector, pick_nearest, verbose, multiply_bloch_phase):
    return _mpb.map_data(d_in_re, d_in_im, n_in, d_out_re, d_out_im, n_out, coord_map, kvector, pick_nearest, verbose, multiply_bloch_phase)

def with_hermitian_epsilon():
    return _mpb.with_hermitian_epsilon()
class mode_solver(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr
    MAX_NWORK = _mpb.mode_solver_MAX_NWORK
    epsilon_CURFIELD_TYPE = _mpb.mode_solver_epsilon_CURFIELD_TYPE
    mu_CURFIELD_TYPE = _mpb.mode_solver_mu_CURFIELD_TYPE
    NUM_FFT_BANDS = _mpb.mode_solver_NUM_FFT_BANDS
    num_bands = property(_mpb.mode_solver_num_bands_get, _mpb.mode_solver_num_bands_set)
    resolution = property(_mpb.mode_solver_resolution_get, _mpb.mode_solver_resolution_set)
    target_freq = property(_mpb.mode_solver_target_freq_get, _mpb.mode_solver_target_freq_set)
    tolerance = property(_mpb.mode_solver_tolerance_get, _mpb.mode_solver_tolerance_set)
    mesh_size = property(_mpb.mode_solver_mesh_size_get, _mpb.mode_solver_mesh_size_set)
    negative_epsilon_ok = property(_mpb.mode_solver_negative_epsilon_ok_get, _mpb.mode_solver_negative_epsilon_ok_set)
    epsilon_input_file = property(_mpb.mode_solver_epsilon_input_file_get, _mpb.mode_solver_epsilon_input_file_set)
    mu_input_file = property(_mpb.mode_solver_mu_input_file_get, _mpb.mode_solver_mu_input_file_set)
    force_mu = property(_mpb.mode_solver_force_mu_get, _mpb.mode_solver_force_mu_set)
    use_simple_preconditioner = property(_mpb.mode_solver_use_simple_preconditioner_get, _mpb.mode_solver_use_simple_preconditioner_set)
    grid_size = property(_mpb.mode_solver_grid_size_get, _mpb.mode_solver_grid_size_set)
    n = property(_mpb.mode_solver_n_get, _mpb.mode_solver_n_set)
    local_N = property(_mpb.mode_solver_local_N_get, _mpb.mode_solver_local_N_set)
    N_start = property(_mpb.mode_solver_N_start_get, _mpb.mode_solver_N_start_set)
    alloc_N = property(_mpb.mode_solver_alloc_N_get, _mpb.mode_solver_alloc_N_set)
    nwork_alloc = property(_mpb.mode_solver_nwork_alloc_get, _mpb.mode_solver_nwork_alloc_set)
    eigensolver_nwork = property(_mpb.mode_solver_eigensolver_nwork_get, _mpb.mode_solver_eigensolver_nwork_set)
    eigensolver_block_size = property(_mpb.mode_solver_eigensolver_block_size_get, _mpb.mode_solver_eigensolver_block_size_set)
    last_parity = property(_mpb.mode_solver_last_parity_get, _mpb.mode_solver_last_parity_set)
    iterations = property(_mpb.mode_solver_iterations_get, _mpb.mode_solver_iterations_set)
    eigensolver_flops = property(_mpb.mode_solver_eigensolver_flops_get, _mpb.mode_solver_eigensolver_flops_set)
    geometry_list = property(_mpb.mode_solver_geometry_list_get, _mpb.mode_solver_geometry_list_set)
    geometry_tree = property(_mpb.mode_solver_geometry_tree_get, _mpb.mode_solver_geometry_tree_set)
    vol = property(_mpb.mode_solver_vol_get, _mpb.mode_solver_vol_set)
    R = property(_mpb.mode_solver_R_get, _mpb.mode_solver_R_set)
    G = property(_mpb.mode_solver_G_get, _mpb.mode_solver_G_set)
    mdata = property(_mpb.mode_solver_mdata_get, _mpb.mode_solver_mdata_set)
    mtdata = property(_mpb.mode_solver_mtdata_get, _mpb.mode_solver_mtdata_set)
    curfield_band = property(_mpb.mode_solver_curfield_band_get, _mpb.mode_solver_curfield_band_set)
    cur_kvector = property(_mpb.mode_solver_cur_kvector_get, _mpb.mode_solver_cur_kvector_set)
    Rm = property(_mpb.mode_solver_Rm_get, _mpb.mode_solver_Rm_set)
    Gm = property(_mpb.mode_solver_Gm_get, _mpb.mode_solver_Gm_set)
    H = property(_mpb.mode_solver_H_get, _mpb.mode_solver_H_set)
    Hblock = property(_mpb.mode_solver_Hblock_get, _mpb.mode_solver_Hblock_set)
    muinvH = property(_mpb.mode_solver_muinvH_get, _mpb.mode_solver_muinvH_set)
    W = property(_mpb.mode_solver_W_get, _mpb.mode_solver_W_set)
    freqs = property(_mpb.mode_solver_freqs_get, _mpb.mode_solver_freqs_set)
    verbose = property(_mpb.mode_solver_verbose_get, _mpb.mode_solver_verbose_set)
    deterministic = property(_mpb.mode_solver_deterministic_get, _mpb.mode_solver_deterministic_set)

    def __init__(self, num_bands, resolution, lat, tolerance, mesh_size, _default_material, deterministic, target_freq, dims, verbose, periodicity, flops, negative_epsilon_ok, epsilon_input_file, mu_input_file, force_mu, use_simple_preconditioner, grid_size, eigensolver_nwork, eigensolver_block_size):
        _mpb.mode_solver_swiginit(self, _mpb.new_mode_solver(num_bands, resolution, lat, tolerance, mesh_size, _default_material, deterministic, target_freq, dims, verbose, periodicity, flops, negative_epsilon_ok, epsilon_input_file, mu_input_file, force_mu, use_simple_preconditioner, grid_size, eigensolver_nwork, eigensolver_block_size))
    __swig_destroy__ = _mpb.delete_mode_solver

    def init(self, p, reset_fields, geometry, _default_material):
        return _mpb.mode_solver_init(self, p, reset_fields, geometry, _default_material)

    def solve_kpoint(self, kpoint):
        return _mpb.mode_solver_solve_kpoint(self, kpoint)

    def using_mu(self):
        return _mpb.mode_solver_using_mu(self)

    def set_parity(self, p):
        return _mpb.mode_solver_set_parity(self, p)

    def set_num_bands(self, nb):
        return _mpb.mode_solver_set_num_bands(self, nb)

    def get_kpoint_index(self):
        return _mpb.mode_solver_get_kpoint_index(self)

    def set_kpoint_index(self, i):
        return _mpb.mode_solver_set_kpoint_index(self, i)

    def get_epsilon(self):
        return _mpb.mode_solver_get_epsilon(self)

    def get_mu(self):
        return _mpb.mode_solver_get_mu(self)

    def get_epsilon_tensor(self, c1, c2, imag, inv):
        return _mpb.mode_solver_get_epsilon_tensor(self, c1, c2, imag, inv)

    def get_material_pt(self, material, p):
        return _mpb.mode_solver_get_material_pt(self, material, p)

    def material_epsmu(self, material, epsmu, epsmu_inv, eps=True):
        return _mpb.mode_solver_material_epsmu(self, material, epsmu, epsmu_inv, eps)

    def mean_epsilon(self, meps, meps_inv, n, d1, d2, d3, tol, r):
        return _mpb.mode_solver_mean_epsilon(self, meps, meps_inv, n, d1, d2, d3, tol, r)

    def randomize_fields(self):
        return _mpb.mode_solver_randomize_fields(self)

    def init_epsilon(self, geometry_in):
        return _mpb.mode_solver_init_epsilon(self, geometry_in)

    def reset_epsilon(self, geometry):
        return _mpb.mode_solver_reset_epsilon(self, geometry)

    def has_mu(self, geometry):
        return _mpb.mode_solver_has_mu(self, geometry)

    def material_has_mu(self, mt):
        return _mpb.mode_solver_material_has_mu(self, mt)

    def curfield_reset(self):
        return _mpb.mode_solver_curfield_reset(self)

    def get_field_size(self):
        return _mpb.mode_solver_get_field_size(self)

    def get_freqs(self):
        return _mpb.mode_solver_get_freqs(self)

    def get_eigensolver_flops(self):
        return _mpb.mode_solver_get_eigensolver_flops(self)

    def get_iterations(self):
        return _mpb.mode_solver_get_iterations(self)

    def get_efield(self, band):
        return _mpb.mode_solver_get_efield(self, band)

    def get_dfield(self, band):
        return _mpb.mode_solver_get_dfield(self, band)

    def get_hfield(self, band):
        return _mpb.mode_solver_get_hfield(self, band)

    def get_bfield(self, band):
        return _mpb.mode_solver_get_bfield(self, band)

    def get_efield_from_dfield(self):
        return _mpb.mode_solver_get_efield_from_dfield(self)

    def get_curfield(self, data):
        return _mpb.mode_solver_get_curfield(self, data)

    def get_curfield_cmplx(self, cdata):
        return _mpb.mode_solver_get_curfield_cmplx(self, cdata)

    def set_curfield(self, data):
        return _mpb.mode_solver_set_curfield(self, data)

    def set_curfield_cmplx(self, cdata):
        return _mpb.mode_solver_set_curfield_cmplx(self, cdata)

    def get_lattice(self, data):
        return _mpb.mode_solver_get_lattice(self, data)

    def get_eigenvectors(self, p_start, p, cdata):
        return _mpb.mode_solver_get_eigenvectors(self, p_start, p, cdata)

    def get_eigenvectors_slice_dims(self, num_bands):
        return _mpb.mode_solver_get_eigenvectors_slice_dims(self, num_bands)

    def set_eigenvectors(self, b_start, cdata):
        return _mpb.mode_solver_set_eigenvectors(self, b_start, cdata)

    def compute_field_energy(self):
        return _mpb.mode_solver_compute_field_energy(self)

    def compute_energy_in_objects(self, objects):
        return _mpb.mode_solver_compute_energy_in_objects(self, objects)

    def get_curfield_type(self):
        return _mpb.mode_solver_get_curfield_type(self)

    def set_curfield_type(self, t):
        return _mpb.mode_solver_set_curfield_type(self, t)

    def get_parity_string(self):
        return _mpb.mode_solver_get_parity_string(self)

    def get_dims(self):
        return _mpb.mode_solver_get_dims(self)

    def set_grid_size(self, gs):
        return _mpb.mode_solver_set_grid_size(self, gs)

    def get_libctl_dimensions(self):
        return _mpb.mode_solver_get_libctl_dimensions(self)

    def set_libctl_dimensions(self, val):
        return _mpb.mode_solver_set_libctl_dimensions(self, val)

    def get_libctl_ensure_periodicity(self):
        return _mpb.mode_solver_get_libctl_ensure_periodicity(self)

    def set_libctl_ensure_periodicity(self, val):
        return _mpb.mode_solver_set_libctl_ensure_periodicity(self, val)

    def set_libctl_geometry_lattice(self, val):
        return _mpb.mode_solver_set_libctl_geometry_lattice(self, val)

    def get_output_k(self):
        return _mpb.mode_solver_get_output_k(self)

    def get_val(self, ix, iy, iz, nx, ny, nz, last_dim_size, data, stride, conjugate):
        return _mpb.mode_solver_get_val(self, ix, iy, iz, nx, ny, nz, last_dim_size, data, stride, conjugate)

    def interp_val(self, p, nx, ny, nz, last_dim_size, data, stride, conjugate):
        return _mpb.mode_solver_interp_val(self, p, nx, ny, nz, last_dim_size, data, stride, conjugate)

    def interp_cval(self, p, nx, ny, nz, last_dim_size, data, stride):
        return _mpb.mode_solver_interp_cval(self, p, nx, ny, nz, last_dim_size, data, stride)

    def interp_eps_inv(self, p):
        return _mpb.mode_solver_interp_eps_inv(self, p)

    def get_epsilon_point(self, p):
        return _mpb.mode_solver_get_epsilon_point(self, p)

    def get_epsilon_inverse_tensor_point(self, p):
        return _mpb.mode_solver_get_epsilon_inverse_tensor_point(self, p)

    def get_energy_point(self, p):
        return _mpb.mode_solver_get_energy_point(self, p)

    def get_field_point(self, p):
        return _mpb.mode_solver_get_field_point(self, p)

    def get_bloch_field_point_(self, field, p):
        return _mpb.mode_solver_get_bloch_field_point_(self, field, p)

    def get_bloch_field_point(self, p):
        return _mpb.mode_solver_get_bloch_field_point(self, p)

    def multiply_bloch_phase(self, cdata=None):
        return _mpb.mode_solver_multiply_bloch_phase(self, cdata)

    def fix_field_phase(self):
        return _mpb.mode_solver_fix_field_phase(self)

    def compute_field_divergence(self):
        return _mpb.mode_solver_compute_field_divergence(self)

    def compute_zparities(self):
        return _mpb.mode_solver_compute_zparities(self)

    def compute_yparities(self):
        return _mpb.mode_solver_compute_yparities(self)

    def compute_group_velocity_component(self, d):
        return _mpb.mode_solver_compute_group_velocity_component(self, d)

    def compute_1_group_velocity_component(self, d, b):
        return _mpb.mode_solver_compute_1_group_velocity_component(self, d, b)

    def compute_1_group_velocity(self, b):
        return _mpb.mode_solver_compute_1_group_velocity(self, b)

    def compute_1_group_velocity_reciprocal(self, b):
        return _mpb.mode_solver_compute_1_group_velocity_reciprocal(self, b)

    def compute_energy_in_dielectric(self, eps_low, eps_high):
        return _mpb.mode_solver_compute_energy_in_dielectric(self, eps_low, eps_high)

    def compute_field_integral(self, field_func):
        return _mpb.mode_solver_compute_field_integral(self, field_func)

    def compute_energy_integral(self, field_func):
        return _mpb.mode_solver_compute_energy_integral(self, field_func)

    def get_dominant_planewave(self, band):
        return _mpb.mode_solver_get_dominant_planewave(self, band)

    def transformed_overlap(self, W, w):
        return _mpb.mode_solver_transformed_overlap(self, W, w)

    def compute_symmetry(self, which_band, W, w):
        return _mpb.mode_solver_compute_symmetry(self, which_band, W, w)

# Register mode_solver in _mpb:
_mpb.mode_solver_swigregister(mode_solver)
cvar = _mpb.cvar


__version__ = (_mpb.cvar.MPB_VERSION_MAJOR, _mpb.cvar.MPB_VERSION_MINOR, _mpb.cvar.MPB_VERSION_PATCH)

from meep.verbosity_mgr import Verbosity
verbosity = Verbosity(_mpb.cvar, 'mpb', 1)

from .solver import (
    MPBArray,
    ModeSolver,
    output_hfield,
    output_hfield_x,
    output_hfield_y,
    output_hfield_z,
    output_bfield,
    output_bfield_x,
    output_bfield_y,
    output_bfield_z,
    output_dfield,
    output_dfield_x,
    output_dfield_y,
    output_dfield_z,
    output_efield,
    output_efield_x,
    output_efield_y,
    output_efield_z,
    output_charge_density,
    output_bpwr,
    output_dpwr,
    output_tot_pwr,
    output_dpwr_in_objects,
    output_poynting,
    output_poynting_x,
    output_poynting_y,
    output_poynting_z,
    output_at_kpoint,
    display_group_velocities,
    display_yparities,
    display_zparities,
    fix_hfield_phase,
    fix_bfield_phase,
    fix_dfield_phase,
    fix_efield_phase,
)

from .mpb_data import (
    MPBData,
)



MPB_VERSION_MAJOR = cvar.MPB_VERSION_MAJOR
MPB_VERSION_MINOR = cvar.MPB_VERSION_MINOR
MPB_VERSION_PATCH = cvar.MPB_VERSION_PATCH




import cmath
import math
import unittest

import numpy as np
import parameterized

import meep as mp


class TestEigCoeffs(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.resolution = 30  # pixels/μm

        cls.dpml = 1.0  # PML thickness
        cls.dsub = 1.0  # substrate thickness
        cls.dpad = 1.0  # padding thickness between grating and PML
        cls.gp = 6.0  # grating period
        cls.gh = 0.5  # grating height
        cls.gdc = 0.5  # grating duty cycle

        cls.sx = cls.dpml + cls.dsub + cls.gh + cls.dpad + cls.dpml
        cls.sy = cls.gp

        cls.cell_size = mp.Vector3(cls.sx, cls.sy, 0)

        # replace anisotropic PML with isotropic Absorber to
        # attenuate parallel-directed fields of oblique source
        cls.abs_layers = [mp.Absorber(thickness=cls.dpml, direction=mp.X)]

        wvl = 0.5  # center wavelength
        cls.fcen = 1 / wvl  # center frequency
        cls.df = 0.05 * cls.fcen  # frequency width

        cls.ng = 1.5
        cls.glass = mp.Medium(index=cls.ng)

        cls.geometry = [
            mp.Block(
                material=cls.glass,
                size=mp.Vector3(cls.dpml + cls.dsub, mp.inf, mp.inf),
                center=mp.Vector3(-0.5 * cls.sx + 0.5 * (cls.dpml + cls.dsub), 0, 0),
            ),
            mp.Block(
                material=cls.glass,
                size=mp.Vector3(cls.gh, cls.gdc * cls.gp, mp.inf),
                center=mp.Vector3(
                    -0.5 * cls.sx + cls.dpml + cls.dsub + 0.5 * cls.gh, 0, 0
                ),
            ),
        ]

    @parameterized.parameterized.expand([(0,), (10.7,)])
    def test_binary_grating_oblique(self, theta):
        # rotation angle of incident planewave
        # counterclockwise (CCW) about Z axis, 0 degrees along +X axis
        theta_in = math.radians(theta)

        # k (in source medium) with correct length (plane of incidence: XY)
        k = mp.Vector3(self.fcen * self.ng).rotate(mp.Vector3(0, 0, 1), theta_in)

        symmetries = []
        eig_parity = mp.ODD_Z
        if theta_in == 0:
            symmetries = [mp.Mirror(mp.Y)]
            eig_parity += mp.EVEN_Y

        def pw_amp(k, x0):
            def _pw_amp(x):
                return cmath.exp(1j * 2 * math.pi * k.dot(x + x0))

            return _pw_amp

        src_pt = mp.Vector3(-0.5 * self.sx + self.dpml, 0, 0)
        sources = [
            mp.Source(
                mp.GaussianSource(self.fcen, fwidth=self.df),
                component=mp.Ez,  # S polarization
                center=src_pt,
                size=mp.Vector3(0, self.sy, 0),
                amp_func=pw_amp(k, src_pt),
            )
        ]

        sim = mp.Simulation(
            resolution=self.resolution,
            cell_size=self.cell_size,
            boundary_layers=self.abs_layers,
            k_point=k,
            default_material=self.glass,
            sources=sources,
            symmetries=symmetries,
        )

        refl_pt = mp.Vector3(-0.5 * self.sx + self.dpml + 0.5 * self.dsub, 0, 0)
        refl_flux = sim.add_flux(
            self.fcen,
            0,
            1,
            mp.FluxRegion(center=refl_pt, size=mp.Vector3(0, self.sy, 0)),
        )

        sim.run(until_after_sources=mp.stop_when_dft_decayed())

        input_flux = mp.get_fluxes(refl_flux)
        input_flux_data = sim.get_flux_data(refl_flux)

        sim.reset_meep()

        sim = mp.Simulation(
            resolution=self.resolution,
            cell_size=self.cell_size,
            boundary_layers=self.abs_layers,
            geometry=self.geometry,
            k_point=k,
            sources=sources,
            symmetries=symmetries,
        )

        refl_flux = sim.add_flux(
            self.fcen,
            0,
            1,
            mp.FluxRegion(center=refl_pt, size=mp.Vector3(0, self.sy, 0)),
        )

        sim.load_minus_flux_data(refl_flux, input_flux_data)

        tran_pt = mp.Vector3(0.5 * self.sx - self.dpml - 0.5 * self.dpad, 0, 0)
        tran_flux = sim.add_flux(
            self.fcen,
            0,
            1,
            mp.FluxRegion(center=tran_pt, size=mp.Vector3(0, self.sy, 0)),
        )

        sim.run(until_after_sources=mp.stop_when_dft_decayed())

        # number of reflected orders
        nm_r = np.floor((self.fcen * self.ng - k.y) * self.gp) - np.ceil(
            (-self.fcen * self.ng - k.y) * self.gp
        )
        if theta_in == 0:
            nm_r = nm_r / 2  # since eig_parity removes degeneracy in y-direction
        nm_r = int(nm_r)

        res = sim.get_eigenmode_coefficients(
            refl_flux, range(1, nm_r + 1), eig_parity=eig_parity
        )
        r_coeffs = res.alpha

        Rsum = 0
        for nm in range(nm_r):
            Rsum += abs(r_coeffs[nm, 0, 1]) ** 2 / input_flux[0]

        # number of transmitted orders
        nm_t = np.floor((self.fcen - k.y) * self.gp) - np.ceil(
            (-self.fcen - k.y) * self.gp
        )
        if theta_in == 0:
            nm_t = nm_t / 2  # since eig_parity removes degeneracy in y-direction
        nm_t = int(nm_t)

        res = sim.get_eigenmode_coefficients(
            tran_flux, range(1, nm_t + 1), eig_parity=eig_parity
        )
        t_coeffs = res.alpha

        Tsum = 0
        for nm in range(nm_t):
            Tsum += abs(t_coeffs[nm, 0, 0]) ** 2 / input_flux[0]

        r_flux = mp.get_fluxes(refl_flux)
        t_flux = mp.get_fluxes(tran_flux)
        Rflux = -r_flux[0] / input_flux[0]
        Tflux = t_flux[0] / input_flux[0]

        print(f"refl:, {Rsum}, {Rflux}")
        print(f"tran:, {Tsum}, {Tflux}")
        print(f"sum:,  {Rsum + Tsum}, {Rflux + Tflux}")

        self.assertAlmostEqual(Rsum, Rflux, places=2)
        self.assertAlmostEqual(Tsum, Tflux, places=2)
        self.assertAlmostEqual(Rsum + Tsum, 1.00, places=2)

    @parameterized.parameterized.expand(
        [(13.2, "real/imag"), (17.7, "complex"), (21.2, "3d")]
    )
    def test_binary_grating_special_kz(self, theta, kz_2d):
        # rotation angle of incident planewave
        # counterclockwise (CCW) about Y axis, 0 degrees along +X axis
        theta_in = math.radians(theta)

        # k (in source medium) with correct length (plane of incidence: XZ)
        k = mp.Vector3(self.fcen * self.ng).rotate(mp.Vector3(0, 1, 0), theta_in)

        symmetries = [mp.Mirror(mp.Y)]

        def pw_amp(k, x0):
            def _pw_amp(x):
                return cmath.exp(1j * 2 * math.pi * k.dot(x + x0))

            return _pw_amp

        src_pt = mp.Vector3(-0.5 * self.sx + self.dpml, 0, 0)
        sources = [
            mp.Source(
                mp.GaussianSource(self.fcen, fwidth=self.df),
                component=mp.Ez,
                center=src_pt,
                size=mp.Vector3(0, self.sy, 0),
                amp_func=pw_amp(k, src_pt),
            )
        ]

        sim = mp.Simulation(
            resolution=self.resolution,
            cell_size=self.cell_size,
            boundary_layers=self.abs_layers,
            k_point=k,
            default_material=self.glass,
            sources=sources,
            symmetries=symmetries,
            kz_2d=kz_2d,
        )

        refl_pt = mp.Vector3(-0.5 * self.sx + self.dpml + 0.5 * self.dsub, 0, 0)
        refl_flux = sim.add_mode_monitor(
            self.fcen,
            0,
            1,
            mp.FluxRegion(center=refl_pt, size=mp.Vector3(0, self.sy, 0)),
        )

        sim.run(until_after_sources=mp.stop_when_dft_decayed())

        input_flux = mp.get_fluxes(refl_flux)
        input_flux_data = sim.get_flux_data(refl_flux)

        sim.reset_meep()

        sim = mp.Simulation(
            resolution=self.resolution,
            cell_size=self.cell_size,
            boundary_layers=self.abs_layers,
            geometry=self.geometry,
            k_point=k,
            sources=sources,
            symmetries=symmetries,
            kz_2d=kz_2d,
        )

        refl_flux = sim.add_mode_monitor(
            self.fcen,
            0,
            1,
            mp.FluxRegion(center=refl_pt, size=mp.Vector3(0, self.sy, 0)),
        )

        sim.load_minus_flux_data(refl_flux, input_flux_data)

        tran_pt = mp.Vector3(0.5 * self.sx - self.dpml - 0.5 * self.dpad, 0, 0)
        tran_flux = sim.add_mode_monitor(
            self.fcen,
            0,
            1,
            mp.FluxRegion(center=tran_pt, size=mp.Vector3(0, self.sy, 0)),
        )

        sim.run(until_after_sources=mp.stop_when_dft_decayed())

        # number of reflected orders
        nm_r = np.ceil(
            (np.sqrt((self.fcen * self.ng) ** 2 - k.z**2) - k.y) * self.gp
        ) - np.floor((-np.sqrt((self.fcen * self.ng) ** 2 - k.z**2) - k.y) * self.gp)
        nm_r = int(nm_r / 2)

        Rsum = 0
        for nm in range(nm_r):
            for S_pol in [False, True]:
                res = sim.get_eigenmode_coefficients(
                    refl_flux,
                    mp.DiffractedPlanewave(
                        [0, nm, 0],
                        mp.Vector3(1, 0, 0),
                        1 if S_pol else 0,
                        0 if S_pol else 1,
                    ),
                )
                r_coeffs = res.alpha
                Rmode = abs(r_coeffs[0, 0, 1]) ** 2 / input_flux[0]
                print(
                    "refl-order:, {}, {}, {}".format("s" if S_pol else "p", nm, Rmode)
                )
                Rsum += Rmode if nm == 0 else 2 * Rmode

        # number of transmitted orders
        nm_t = np.ceil((np.sqrt(self.fcen**2 - k.z**2) - k.y) * self.gp) - np.floor(
            (-np.sqrt(self.fcen**2 - k.z**2) - k.y) * self.gp
        )
        nm_t = int(nm_t / 2)

        Tsum = 0
        for nm in range(nm_t):
            for S_pol in [False, True]:
                res = sim.get_eigenmode_coefficients(
                    tran_flux,
                    mp.DiffractedPlanewave(
                        [0, nm, 0],
                        mp.Vector3(1, 0, 0),
                        1 if S_pol else 0,
                        0 if S_pol else 1,
                    ),
                )
                t_coeffs = res.alpha
                Tmode = abs(t_coeffs[0, 0, 0]) ** 2 / input_flux[0]
                print(
                    "tran-order:, {}, {}, {}".format("s" if S_pol else "p", nm, Tmode)
                )
                Tsum += Tmode if nm == 0 else 2 * Tmode

        r_flux = mp.get_fluxes(refl_flux)
        t_flux = mp.get_fluxes(tran_flux)
        Rflux = -r_flux[0] / input_flux[0]
        Tflux = t_flux[0] / input_flux[0]

        print(f"refl:, {Rsum}, {Rflux}")
        print(f"tran:, {Tsum}, {Tflux}")
        print(f"sum:,  {Rsum + Tsum}, {Rflux + Tflux}")

        self.assertAlmostEqual(Rsum, Rflux, places=2)
        self.assertAlmostEqual(Tsum, Tflux, places=2)
        self.assertAlmostEqual(Rsum + Tsum, 1.00, places=2)


if __name__ == "__main__":
    unittest.main()



# 1d simulation of a plane wave propagating through a Kerr medium
# and generating the third-harmonic frequency component.
import argparse

import meep as mp


def main(args):

    sz = 100  # size of cell in z direction
    fcen = 1 / 3.0  # center frequency of source
    df = fcen / 20.0  # frequency width of source
    amp = args.amp  # amplitude of source
    k = 10**args.logk  # Kerr susceptibility
    dpml = 1.0  # PML thickness

    # We'll use an explicitly 1d simulation.  Setting dimensions=1 will actually
    # result in faster execution than just using two no-size dimensions.  However,
    # in this case Meep requires us to use E in the x direction (and H in y),
    # and our one no-size dimension must be z.
    dimensions = 1
    cell = mp.Vector3(0, 0, sz)
    pml_layers = mp.PML(dpml)
    resolution = 20

    # to put the same material in all space, we can just set the default material
    # and pass it to the Simulation constructor
    default_material = mp.Medium(index=1, chi3=k)

    sources = mp.Source(
        mp.GaussianSource(fcen, fwidth=df),
        component=mp.Ex,
        center=mp.Vector3(0, 0, -0.5 * sz + dpml),
        amplitude=amp,
    )

    # frequency range for flux calculation
    nfreq = 400
    fmin = fcen / 2.0
    fmax = fcen * 4

    sim = mp.Simulation(
        cell_size=cell,
        geometry=[],
        sources=[sources],
        boundary_layers=[pml_layers],
        default_material=default_material,
        resolution=resolution,
        dimensions=dimensions,
    )

    # trans = sim.add_flux(0.5 * (fmin + fmax), fmax - fmin, nfreq,
    #                      mp.FluxRegion(mp.Vector3(0, 0, 0.5*sz - dpml - 0.5)))
    trans1 = sim.add_flux(
        fcen, 0, 1, mp.FluxRegion(mp.Vector3(0, 0, 0.5 * sz - dpml - 0.5))
    )
    trans3 = sim.add_flux(
        3 * fcen, 0, 1, mp.FluxRegion(mp.Vector3(0, 0, 0.5 * sz - dpml - 0.5))
    )

    sim.run(
        until_after_sources=mp.stop_when_fields_decayed(
            50, mp.Ex, mp.Vector3(0, 0, 0.5 * sz - dpml - 0.5), 1e-6
        )
    )

    # sim.display_fluxes(trans)
    print(
        f"harmonics:, {k}, {amp}, {mp.get_fluxes(trans1)[0]}, {mp.get_fluxes(trans3)[0]}"
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-amp", type=float, default=1.0, help="amplitude of source")
    parser.add_argument(
        "-logk", type=float, default=0, help="logarithm of Kerr susceptibility"
    )
    args = parser.parse_args()
    main(args)
